from dataclasses import dataclass
from typing import Optional

from tidychef.against.implementations.base import BaseValidator
from tidychef.models.source.cell import Cell


@dataclass
class LengthValidator(BaseValidator):
    """
    A class to return bool (valid or not) when
    called with a single instance of a tidychef
    Cell object.
    """

    least: Optional[int] = None
    most: Optional[int] = None

    def __post_init__(self):
        assert not all([self.least is None, self.most is None]), (
            "To use a length validator you must provide at least one "
            "keyword argument from least= and most="
        )
        if self.least and self.most:
            assert (
                self.most > self.least
            ), f"""
                Your most value needs to be bigger than least,
                you have provided:
                                            
                least: {self.least}
                most: {self.most}
            """
        if self.least:
            assert isinstance(self.least, int), "You least value must be an integer."
        if self.most:
            assert isinstance(self.most, int), "You most value must be an integer."

    def __call__(self, cell: Cell) -> bool:
        """
        Is the length of the value property of the Cell
        between the minimum and maximum length

        :param cell: A single tidychef Cell object.
        :return: bool, is it valid or not
        """
        if self.least and self.most:
            return len(cell.value) >= self.least and len(cell.value) <= self.most
        elif self.least:
            return len(cell.value) >= self.least
        else:
            return len(cell.value) <= self.most

    def msg(self, cell: Cell) -> str:
        """
        Provide a contextually meaningful
        message to the user where cell
        is not numeric

        :param cell: A single tidychef Cell object.
        :return: A contextual message
        """
        if self.least and self.most:
            return f'The length of cell value "{cell.value}" is not between {self.least} and {self.most} in length.'
        elif self.least:
            return f'The length of cell value "{cell.value}" is not above the minimum length of {self.least}.'
        else:
            return f'The length of cell value "{cell.value}" is not below the maximum length of {self.most}.'



"""
Common filters.
"""

from dataclasses import dataclass

from tidychef.models.source.cell import Cell


@dataclass
class ContainsString:
    """
    A filter than when given a string, filters cells based on whether
    that string appears in the cell value.
    """

    substr: str

    @property
    def explain(self) -> str:
        return f"Contains string: {self.substr}"

    def __call__(self, cell: Cell):
        return self.substr in cell.value


@dataclass
class NotContainsString:
    """
    A filter than when given a string, filters cells based on whether
    that string does not appear in the cell value.
    """

    substr: str

    @property
    def explain(self) -> str:
        return f"Does not contain string: {self.substr}"

    def __call__(self, cell: Cell):
        return self.substr not in cell.value


class IsNumeric:
    """
    The value of the cell is numerical
    """

    @property
    def explain(self) -> str:
        return "Cell value is numeric"

    def __call__(self, cell: Cell):
        return cell.value.strip().isnumeric()


is_numeric = IsNumeric()


class IsNotNumeric:
    """
    The value of the cell is numerical
    """

    @property
    def explain(self) -> str:
        return "Cell value is not numeric"

    def __call__(self, cell: Cell):
        return not cell.value.strip().isnumeric()


is_not_numeric = IsNotNumeric()



from typing import Callable, Dict

from tidychef.exceptions import (
    BadConditionalResolverError,
    HorizontalConditionalHeaderError,
)
from tidychef.models.source.cell import Cell, VirtualCell

from ..base import BaseLookupEngine


class HorizontalCondition(BaseLookupEngine):
    """
    A lookup engine to populate the contents of a column based
    on the values resolved for the other columns on the row.
    """

    def __init__(
        self,
        label: str,
        resolver: Callable[[Dict[str, str]], str],
        priority: int = 0,
        table: str = "Unnamed Table",
    ):
        """
        A lookup engine to populate the contents of a column based
        on the values resolved for the other columns on the row.

        :param label: The label of the column informed
        by this lookup engine.
        :param resolver: The callable that will create the new column value
        :param priority: The priority used when resolving multiple horizontal
        conditions, 0 is highest priority and the default.
        :param table: the name of the table data is being extracted from
        """
        self.label = label
        self.resolver = resolver
        self.priority = priority
        self.table = table

    def resolve(self, _: Cell, cells_on_row: Dict[str, str]) -> VirtualCell:
        """
        For a given observation row (as denoted by the unused Cell argument),
        resolve the

        :param _: Unused tidychef cell object representing the observation in
        question. Required to match api signature used by other look engines.
        :param cells_on_row: Dictionary containing contents of other
        columns already resolve against the observation cell.
        :return: The value for the column.
        """

        if not isinstance(cells_on_row, dict):
            raise BadConditionalResolverError(
                f"""
                Issue encountered when processing table: "{self.table}".

                A condition resolver should take an argument of type:
                Dict[str, str] and return type str.
                                              
                The resolver for {self.label} is incorrect, it has
                input_type: {type(cells_on_row)}
                """
            )

        try:
            column_value = self.resolver(cells_on_row)
            if not isinstance(column_value, str):
                raise BadConditionalResolverError(
                    f"""
                    Issue encountered when processing table: "{self.table}".

                    A condition resolver should take an argument of type:
                    Dict[str, str] and return type str.
                                                    
                    The resolver for {self.label} is incorrect, it has:

                    return type: {type(column_value)}
                    return value: {column_value} 
                    """
                )
            return VirtualCell(value=column_value)

        except KeyError as err:
            raise HorizontalConditionalHeaderError(
                f"""
                Issue encountered when processing table: "{self.table}".

                Unable to resolve lookup for "{self.label}".
                                               
                The column header key "{err.args[0]}" was specified in:
                the condition but is not (yet?) present on the resolved row:
                
                Header Keys: {cells_on_row.keys()}

                Note - please be aware of ordering when using horizontal conditions
                that interact (i.e where condition 2 required condition 1 to be
                resolved first)

                The priority= keyword can be used with the Column.horizontal_condition()
                constructor where conditionals must be sequenced. The lower the priority
                the sooner the condition is executed.
                """
            ) from err
        except Exception as err:
            raise err



from __future__ import annotations

import csv
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import tabulate
from IPython.core.display import HTML, display

from tidychef.column.base import BaseColumn
from tidychef.exceptions import DroppingNonColumnError, MisalignedHeadersError
from tidychef.lookup.engines.horizontal_condition import HorizontalCondition
from tidychef.models.source.table import LiveTable
from tidychef.notebook.ipython import in_notebook
from tidychef.notebook.preview.html.tidy_data import tidy_data_as_html_table_string
from tidychef.output.base import BaseOutput
from tidychef.utils.decorators import dontmutate



class TidyData(BaseOutput):
    def __init__(
        self,
        observations: LiveTable,
        *columns,
        obs_apply: Callable = None,
        drop: Optional[List[str]] = None,
    ):
        """
        A class to generate a basic representation of the
        data as tidy data.

        :param observations: The cell selection representing observations.
        :param *columns: 1-n Columns to resolve against the observations.
        :param obs_apply: Callable to make changes to values in the observation column.
        :param drop: Columns by label to drop after cells have been resolved.
        """

        assert (
            len(columns) > 0
        ), "You need to pass at least one column like object to TidyData"

        assert (
            observations.label is not None
        ), """
                You must have labelled your observation selection before passing
                it to the TidyData constructor.

                You can do so with the .label_as() method.
            """

        self.observations = observations
        self.columns: List[BaseColumn] = columns
        self.drop = drop if drop else []
        self.obs_apply = obs_apply

        # Don't transform until told to, but once we have
        # only do it once.
        self._data = None

    def __get_representation(self):  # pragma: no cover
        """
        Representation logic shared by __str__ and __repr__
        """
        if not self._data:
            self._transform()

        header_row = self._data[0]
        data_rows = self._data[1:]

        # If we're in a notebook, create a nice html display
        if in_notebook():
            display(HTML(tidy_data_as_html_table_string(self._data)))
            return ""

        # Else return something that'll make sense in a terminal
        return tabulate.tabulate(data_rows, headers=header_row)

    def __repr__(self):  # pragma: no cover
        return self.__get_representation()

    def __str__(self):  # pragma: no cover
        return self.__get_representation()

    def __len__(self):
        self._transform()
        return len(self._data)

    def to_dict(self) -> dict:
        """
        Outputs the TidyData as a pandas style dictionary, i.e

        {
            column1: [column1value1, column1value2, column1value3],
            column2: [column2value1, column2value2, column2value3],
            etc
        }
        """
        self._transform()
        output_dict: Dict[str, List[str]] = {}

        count = 0
        translater: Dict[i, str] = {}

        for column_name in self._data[0]:
            output_dict[column_name] = []
            translater[count] = column_name
            count += 1

        for row in self._data[1:]:
            for i, item in enumerate(row):
                output_dict[translater[i]].append(item)

        return output_dict

    @staticmethod
    def from_tidy(*tidy_data_objects: TidyData) -> TidyData:
        """
        Creates a class:TidyData object from multiple class:TidyData objects
        provided in the form:

        TidyData.from_tidy(tidydata1, tidydata2, tidydata3)

        :param *tidy_data_objects: 1-n populated tidy data objects that we
        want to join together.
        """
        return TidyData.from_tidy_list(list(tidy_data_objects))

    @staticmethod
    def from_tidy_list(tidy_data_objects: List[TidyData]) -> TidyData:
        """
        Creates a class:TidyData object from a list of class:TidyData objects
        provided in the form:

        TidyData.from_tidy_list([tidydata1, tidydata2, tidydata3])

        :param tidy_data_objects: A list of 1-n populated tidy data objects
        that we want to join together.
        """

        for tidy_data_object in tidy_data_objects:
            assert isinstance(
                tidy_data_object, TidyData
            ), """
            Only objects of type TidyData can be passed into
            TidyData.from_many().
            """

        assert (
            len(tidy_data_objects) > 1
        ), """
            You need to pass 2 or more objects of class TidyData
            into to join multiple TidyData sources into one.
        """

        tidy_data = tidy_data_objects[0]
        for remaining_tidy_data in tidy_data_objects[1:]:
            tidy_data += remaining_tidy_data

        return tidy_data

    @dontmutate
    def __add__(self, other_tidy_data: TidyData):
        # Make sure all transforms have happened
        self._transform()
        other_tidy_data._transform()

        # Error if we're joining two TidyData objects
        # with different headers
        if self._data[0] != other_tidy_data._data[0]:
            raise MisalignedHeadersError(
                f"""
                You are attempting to sum two tidy data
                outputs but they do not have the same
                column headers.

                TidyData1 headers:
                {self._data[0]}

                TidyData2 headers:
                {other_tidy_data._data[0]}
            """
            )

        # Since the headers match, join all but the header
        # row from the new source
        self._data += other_tidy_data._data[1:]
        return self

    def _transform(self):
        """
        Uses the column relationships defined to create an
        internal (to this class) representation of tidy data.

        This representation is in the form of:

        [
            [header1,header2,header3],

            [observation1, column2value1, column3value1]

            ...etc...
        ]

        Once this method has been ran, this internal
        representation can be accessed via "._data".

        This result is cached so future calls to transform
        will not repopulate this attribute and will be
        ignored.
        """
        if not self._data:
            grid = []
            drop_count = 0

            # Observations always has the table name so
            # spread it around to cover those columns that
            # are created via static methods.
            for column in self.columns:
                column._table == self.observations._name

            # We need to carefully construct our lists of lists
            # such that the HorizontalCondition columns are created last
            # Ordering will be restored at the end via this list.
            ordered_column_headers = [self.observations.label] + [
                x.label for x in self.columns
            ]

            # Obs label
            if self.observations.label in self.drop:
                unordered_header_row = []
                drop_count += 1
            else:
                unordered_header_row = [self.observations.label]

            # Standard Column labels
            for column in [
                x for x in self.columns if not isinstance(x.engine, HorizontalCondition)
            ]:
                if column.label not in self.drop:
                    unordered_header_row.append(column.label)
                else:
                    drop_count += 1

            # Horizontal Condition
            for column in [
                x for x in self.columns if isinstance(x.engine, HorizontalCondition)
            ]:
                if column.label not in self.drop:
                    unordered_header_row.append(column.label)
                else:
                    drop_count += 1

            # Conditional column labels
            header_row = []
            for col in ordered_column_headers:
                if col in unordered_header_row:
                    header_row.append(col)

            grid.append(header_row)

            # If user has opted to drop a column that does
            # not exist, we need to tell them.
            if drop_count != len(self.drop):
                raise DroppingNonColumnError(
                    f"""
                    You're attempting to drop one or more columns that
                    do not exist in the data.

                    You're dropping: {self.drop}

                    Columns are: {[x.label for x in self.columns]} 
                    """
                )

            for observation in self.observations:
                row_as_dict = {}

                # note we ALWAYS want values in the column_value_dict
                # regardless of whether we're dropping the column
                column_value_dict: Dict[str, str] = {}
                column_value_dict[self.observations.label] = observation.value

                if self.observations.label not in self.drop:
                    row_as_dict[self.observations.label] = (
                        observation.value
                        if not self.obs_apply
                        else self.obs_apply(observation.value)
                    )

                # Resolve the standard columns first
                standard_columns = [
                    x
                    for x in self.columns
                    if not isinstance(x.engine, HorizontalCondition)
                ]
                for column in standard_columns:
                    col_value = column.resolve_column_cell_from_obs_cell(
                        observation
                    ).value
                    column_value_dict[column.label] = col_value
                    if column.label not in self.drop:
                        row_as_dict[column.label] = col_value

                # Now we know the standard column values, resolve the
                # horizontal conditions
                condition_columns = [
                    x for x in self.columns if isinstance(x.engine, HorizontalCondition)
                ]
                priorities = set([x.engine.priority for x in condition_columns])
                for i in priorities:
                    for column in condition_columns:
                        if column.engine.priority == i:
                            col_value = column.resolve_column_cell_from_obs_cell(
                                observation, column_value_dict
                            ).value

                            column_value_dict[column.label] = col_value
                            if column.label not in self.drop:
                                row_as_dict[column.label] = col_value

                # Order for output
                line = []
                for col in ordered_column_headers:
                    value_to_output = row_as_dict.get(col, None)
                    if value_to_output is not None:
                        line.append(value_to_output)

                grid.append(line)

            self._data = grid

    def to_csv(
        self,
        path: Union[str, Path],
        write_headers=True,
        write_mode="w",
        **kwargs,
    ):
        """
        Output the TidyData to a csv file.

        This method wraps the standard csv python library,
        the **kwargs provided here are passed
        through to the csv.csvwriter() constructor.
        https://docs.python.org/3/library/csv.html

        :param path: The location we want to output the
        csv to.
        :param write_headers: Whether or not to include the
        headers when writing to csv.
        :param write_mode: The mode with which to open
        the python file object. Defaults to "w".
        """
        if not self._data:
            self._transform()

        if not isinstance(path, (Path, str)):
            raise ValueError(
                "To output to a file you must provide a pathlib.Path object or a str"
            )

        if isinstance(path, str):
            path = Path(path)

        if not path.parent.exists():
            raise FileNotFoundError(
                f'The specified output directory "{path.parent.absolute()}" for the file "{path.name}" does not exist.'
            )

        with open(path, write_mode) as csvfile:
            tidywriter = csv.writer(csvfile, **kwargs)
            for i, row in enumerate(self._data):
                if i == 0 and not write_headers:
                    continue
                tidywriter.writerow(row)

    def drop_duplicates(
        self,
        print_duplicates: bool = False,
        csv_duplicate_path: Union[str, Path] = None,
    ):
        """
        Drop duplicates from our tidydata.

        :param print_duplicates: Do we want a human friendly report showing
        each duplicate row that has been dropped.
        :param csv_duplicate_path: Path to output a csv containing each duplicate
        row.
        """
        self._transform()
        unique = []

        non_unique = []
        lines = [
            "Removed duplicate instances of the following row(s):",
            "-----------------------------------------------------",
        ]

        for row in self._data:
            if row not in unique:
                unique.append(row)
            else:
                if print_duplicates:  # pragma: no cover
                    lines.append(",".join(row))
                if csv_duplicate_path:
                    non_unique.append(row)

        if print_duplicates:  # pragma: no cover
            for line in lines:
                print(line)

        if csv_duplicate_path:
            if not isinstance(csv_duplicate_path, Path):
                csv_duplicate_path = Path(csv_duplicate_path)

            with open(csv_duplicate_path, "w") as csvfile:
                duplicates_writer = csv.writer(csvfile)
                for row in non_unique:
                    duplicates_writer.writerow(row)

        self._data = unique
        return self



"""
Holds the code that defines the local csv reader.
"""

import csv
from pathlib import Path
from typing import Callable, Optional, Union

from tidychef.acquire.base import BaseReader
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import Table
from tidychef.selection.csv.csv import CsvSelectable
from tidychef.selection.selectable import Selectable
from tidychef.utils import fileutils

from ..base import BaseReader
from ..main import acquirer


def local(
    source: Union[str, Path],
    selectable: Selectable = CsvSelectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    **kwargs
) -> CsvSelectable:
    """
    Read data from a Path (or string representing a path)
    present on the same machine where tidychef is running.

    This local csv reader uses csv.reader() from the standard
    python library. Keyword arguments passed into this function
    are propagated through to csv.reader().
    https://docs.python.org/3/library/csv.html

    :param source: A Path object or a string representing a path
    :param selectable: A class that implements tidychef.selection.selectable.Selectable of an inheritor of. Default is CsvSelectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of LocalCsvReader.parse() as an argument.
    :return: A single populated Selectable of type as specified by selectable param
    """

    return acquirer(
        source,
        LocalCsvReader(),
        selectable,
        pre_hook=pre_hook,
        post_hook=post_hook,
        **kwargs
    )


class LocalCsvReader(BaseReader):
    """
    A reader to lead in a source where that source is a locally
    held csv file.
    """

    def parse(
        self, source: Union[str, Path], selectable: Selectable = CsvSelectable, **kwargs
    ) -> CsvSelectable:
        """
        Parse the provided source into a Selectable. Unless overridden the
        selectable is of type CsvSelectable.

        Optional **kwargs are propagated to the csv.reader() method.

        :param source: A Path or str representing a path indicating a local file
        :param selectable: The selectable type to be returned.
        :return: A list of type as specified by param selectable.
        """

        assert isinstance(
            source, (str, Path)
        ), """
                The source you're passing to acquire.csv.local() needs to
                be either a Path object or a string representing such.
                """

        source: Path = fileutils.ensure_existing_path(source)

        table = Table()
        with open(source, "r", encoding="utf8") as csv_file:
            file_content = csv.reader(csv_file, **kwargs)

            for y_index, row in enumerate(file_content):
                for x_index, cell_value in enumerate(row):
                    table.add_cell(Cell(x=x_index, y=y_index, value=str(cell_value)))

        return selectable(table, source=source)



from .badparams import (
    AmbiguousWaffleError,
    BadExcelReferenceError,
    BadShiftParameterError,
    CardinalDeclarationWithOffset,
    OutputPassedToPreview,
    ReferenceOutsideSelectionError,
    ReversedExcelRefError,
    UnknownDirectionError,
    WithinAxisDeclarationError,
)
from .cells import (
    CellValidationError,
    InvlaidCellPositionError,
    NonExistentCellComparissonError,
)
from .common import (
    BadConditionalResolverError,
    CellsDoNotExistError,
    DroppingNonColumnError,
    FileInputError,
    HorizontalConditionalHeaderError,
    ImpossibleLookupError,
    InvalidCellObjectError,
    LoneValueOnMultipleCellsError,
    MisalignedHeadersError,
    MissingLabelError,
    OutOfBoundsError,
    UnalignedTableOperation,
    ZeroAcquiredTablesError,
)
from .lookups import AmbiguousLookupError, FailedLookupError, MissingDirectLookupError



from dataclasses import dataclass
from typing import List

from tidychef.against.implementations.base import BaseValidator
from tidychef.models.source.cell import Cell


@dataclass
class ItemsValidator(BaseValidator):
    """
    A class to return bool (valid or not) when
    called with a single instance of a tidychef
    Cell object.
    """

    items: List[str]

    def __call__(self, cell: Cell) -> bool:
        """
        Does the value property of the Cell
        appear in the items list

        :param cell: A single tidychef Cell object.
        :return: bool, is it valid or not
        """
        return cell.value in self.items

    def msg(self, cell: Cell) -> str:
        """
        Provide a contextually meaningful
        message to the user where cell
        value not on the provided items list

        :param cell: A single tidychef Cell object.
        :return: A contextual message
        """

        # dont output the full list if its
        # impractically long.
        if len(self.items) < 11:
            return f'"{cell.value}" not in list: {self.items}'
        else:
            return f'"{cell.value}" not in list.'



from __future__ import annotations

from dataclasses import dataclass

from tidychef.exceptions import CardinalDeclarationWithOffset
from tidychef.utils.decorators import dontmutate


class BaseDirection:
    ...


@dontmutate
@dataclass
class Direction(BaseDirection):
    """
    A class representing a cardinal direction.

    :param x: The horizontal offset. For example:
    1 = one space, right, -1 = one position left.
    :param y: The vertical offset. For example:
    1 = one space down, -1 = one position left.
    :param name: The name of the direction
    :param _locked: Can this directions offset be overwritten.
    """

    x: int
    y: int
    name: str
    _locked: bool = False

    @property
    def offset_as_str(self) -> str:
        """
        Return a plain english description of a direction
        offset so we can give the user some kind of plain
        English feedback.
        """
        if self.is_right:
            offset = self.x
        if self.is_left:
            offset = 0 - self.x
        if self.is_downwards:
            offset = self.y
        if self.is_upwards:
            offset = 0 - self.y

        return f"{self.name}({offset})"

    @property
    def is_upwards(self) -> bool:
        """
        Is direction.name one of ["up", "above"]
        """
        return self.name in ["up", "above"]

    @property
    def is_downwards(self) -> bool:
        """
        Is direction.name one of ["down", "below"]
        """
        return self.name in ["down", "below"]

    @property
    def is_left(self) -> bool:
        """
        Is direction.name "left"
        """
        return self.name == "left"

    @property
    def is_right(self) -> bool:
        """
        Is direction.name "right"
        """
        return self.name == "right"

    @property
    def is_horizontal(self) -> bool:
        """
        Is left or right
        """
        return self.is_right or self.is_left

    @property
    def is_vertical(self) -> bool:
        """
        Is up, down, above or below
        """
        return self.is_upwards or self.is_downwards

    def inverted(self) -> Direction:
        """
        Returns a direction that is the direct inversion of
        this direction.

        for example:

        up(2).inverted() == down(2)
        left(8).inverted() == right(8)
        """

        if self.is_upwards:
            offset = 0 - self.y
            return down(offset)

        if self.is_downwards:
            offset = 0 + self.y
            return up(offset)

        if self.is_right:
            offset = 0 + self.x
            return left(offset)

        if self.is_left:
            offset = 0 - self.x
            return right(offset)

    def _confirm_pristine(self):
        """
        There are some scenarios where we want to disallow
        the passing in of directional offsets.
        """
        if self._locked:
            raise CardinalDeclarationWithOffset(
                """
                You cannot pass an offset into a direction
                in the context of declaring an absolute direction.
            
                i.e .expand(up) or .fill(up) is ok, expand(up(1)) or
                fill(up(1) is not.
                """
            )

    def __call__(self, relative_change: int):
        """
        Passing in an integer a parameter enables us to
        declare shifts along the xy axis (where applicable)
        without dipping into raw xy positioning
        e,g: right(3), down(5).

        For example:

        right (default) == (x:1, y:0)
        So equates to a representation of one positive
        positional movement on the x axis.
        So right(3) == (x:4, y:0)

        left == (x:-1, y:0)
        so left(3) == (x:-4: y:0)

        :param relative_change: the amount to extend the
        positional integer by.
        """

        # Don't allow continual overwriting of a directional offset
        # i.e right(3)(4) would be a mess.
        if self._locked:
            raise Exception(
                "You cannot modify a cardinal directions offset value more than once"
            )

        if self.is_horizontal:
            if self.x == -1:
                x = -relative_change
                y = self.y
            elif self.x == 1:
                x = relative_change
                y = self.y
        else:
            if self.y == -1:
                y = -relative_change
                x = self.x
            elif self.y == 1:
                y = relative_change
                x = self.x
        return Direction(x, y, self.name, _locked=True)


up = Direction(0, -1, "up")
above = Direction(0, -1, "above")
down = Direction(0, 1, "down")
below = Direction(0, 1, "below")
right = Direction(1, 0, "right")
left = Direction(-1, 0, "left")



class NonExistentCellComparissonError(Exception):
    """
    User is trying to to make a positional comparison between an
    existant cell (i.e a cell parsed from a tabulated source) and
    a virtual cell created by a constant or other external input.
    """

    def __init__(self, msg):
        self.msg = msg


class InvlaidCellPositionError(Exception):
    """
    Raised where a cell has a value for one positional index but no the other.

    Examples:
    y has a value but x does not
    x has a value but y does not
    """

    def __init__(self, msg):
        self.msg = msg


class CellValidationError(Exception):
    """
    Raised where we're validated a call value and
    found it to me invalid.
    """

    def __init__(self, msg):
        self.msg = msg



from tidychef.models.source.cell import Cell, VirtualCell

from ..base import BaseLookupEngine


class Constant(BaseLookupEngine):
    """
    A class to resolve a direct lookup between
    an observation cell and a single constant
    value.
    """

    def __init__(self, label: str, value: str):
        """
        A class to resolve a direct lookup between
        an observation cell and a single constant
        value.

        :param label: The label of the column informed
        by this lookup engine.
        :param value: The constant value we want this
        lookup engine to return.
        """
        self.cell = VirtualCell(value=value)
        self.label = label

    def resolve(self, _: Cell):
        """
        Regardless of the observation cell,
        return the constant cell.

        :param _: Unused Cell object required to keep api
        signature in keeping with the other engines
        """
        return self.cell



"""
TODO - works but messy, rewrite for clarity once we've a comprehensive test
suite in place.
"""

from typing import List

from tidychef.models.source.table import LiveTable
from tidychef.utils import cellutils

from ..boundary import Boundary
from .components import HtmlCell, SelectionKeys
from .constants import (
    BORDER_CELL_COLOUR,
    BORDER_CELL_SECONDARY_COLOUR,
    INLINE_CSS,
    NO_COLOUR,
    WARNING_COLOUR,
)


def get_preview_table_as_html(
    selections: List[LiveTable],
    bounded: str,
    show_excel: bool = True,
    show_xy: bool = False,
    border_cells: str = BORDER_CELL_COLOUR,
    blank_cells: str = NO_COLOUR,
    warning_colour: str = WARNING_COLOUR,
    border_cell_secondary_colour: str = BORDER_CELL_SECONDARY_COLOUR,
    multiple_selection_warning: bool = True,
    selection_boundary: bool = False,
) -> str:
    """ """

    selection_keys = SelectionKeys()
    for selection in selections:
        selection: LiveTable

        # If the selection is pristine, someone is just
        # previewing the data prior to selections
        if not selection.selections_made() and selection_boundary is False:
            continue
        selection_keys.add_selection_key(selection)

    boundary = Boundary(
        selections, bounded=bounded, selection_boundary=selection_boundary
    )
    all_cells: LiveTable = selections[0].pcells
    all_cells = [
        cell
        for cell in all_cells
        if cell.x >= boundary.leftmost_point
        and cell.x <= boundary.rightmost_point
        and cell.y >= boundary.highest_point
        and cell.y <= boundary.lowest_point
    ]

    html_cell_rows = []
    last_y = boundary.highest_point
    row = []
    show_warning = False

    # Add table headers as needed.
    if show_xy or show_excel:
        if show_xy:
            row.append(HtmlCell("x/y", border_cell_secondary_colour))
            if show_excel:
                row.append(HtmlCell("", border_cell_secondary_colour))
            for i in range(boundary.leftmost_point, boundary.rightmost_point + 1):
                row.append(HtmlCell(i, border_cell_secondary_colour))
            html_cell_rows.append(row)
            row = []

        if show_excel:
            if show_xy:
                row.append(HtmlCell("", border_cell_secondary_colour))
            row.append(HtmlCell("", border_cells))
            for i in range(boundary.leftmost_point, boundary.rightmost_point + 1):
                letters = cellutils.x_to_letters(i)
                row.append(HtmlCell(letters, border_cells))
            html_cell_rows.append(row)
            row = []

        if show_xy:
            row.append(HtmlCell(last_y, border_cell_secondary_colour))
        if show_excel:
            row.append(HtmlCell(last_y + 1, border_cells))

    # Add cell rows, including xy and excel if so indicated
    for cell in all_cells:
        if cell.y != last_y:
            html_cell_rows.append(row)
            row = []
            if show_xy:
                row.append(HtmlCell(cell.y, border_cell_secondary_colour))
            if show_excel:
                row.append(HtmlCell(cell.y + 1, border_cells))
            last_y = cell.y

        found = 0
        for selection_key in selection_keys:
            if selection_key.matches_xy_of_cell(cell):
                found += 1
                colour = selection_key.colour

        if found == 1:
            row.append(HtmlCell(cell.value, colour=colour))
        elif found > 1:
            if multiple_selection_warning:
                row.append(HtmlCell(cell.value, colour=warning_colour))
                show_warning = True
            else:
                row.append(HtmlCell(cell.value, colour=colour))
        else:
            row.append(HtmlCell(cell.value, blank_cells))

    # final row
    html_cell_rows.append(row)

    # ---------------------
    # Create html key table

    if show_warning:
        key_table_html = f"""
            <tr>
                <td style="background-color:{warning_colour}">Cell Appears in Multiple Selections</td>
            <tr>
        """
    else:
        key_table_html = ""

    for selection_key in selection_keys:
        key_table_html += selection_key.as_html()

    # ---------------------
    # Create html cell rows

    cell_table_row_html = ""
    for row in html_cell_rows:
        row: List[HtmlCell]
        row_as_html = "".join([x.as_html() for x in row])
        cell_table_row_html += f"<tr>{row_as_html}</tr>\n"

    # ------------
    # Create tables

    return f"""
    <html>
        {INLINE_CSS}
            <table>
                {key_table_html}
            </table>

            <body>
                <h2>{selections[0].name}</h2>
                <table>
                    {cell_table_row_html}
                </table>
            </body>
            <br>
        </html>
    """



from typing import Dict, List, Optional

from tidychef import datafuncs as dfc
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import LiveTable


class Boundary:
    """
    Class to calculate and help manage the boundary of
    the cells we wish to display in this preview.
    """

    def __init__(
        self,
        selections: List[LiveTable],
        bounded: str = None,
        selection_boundary: bool = False,
    ):
        """
        Created a boundary to calculate the size of the
        preview that should be generated.

        :param selections: A list of all current selections
        :param bounded: An optional excel style reference to
        constrain the size of the previewed table.
        """

        assert not all([selection_boundary, bounded])
        if selection_boundary:
            assert len(selections) == 1
            selection = selections[0].cells
        else:
            selection = selections[0].pcells

        self.bounded = bounded
        if bounded is None:
            self.max_selected_x: int = dfc.maximum_x_offset(selection)
            self.max_selected_y: int = dfc.maximum_y_offset(selection)
            self.min_selected_x: int = dfc.minimum_x_offset(selection)
            self.min_selected_y: int = dfc.minimum_y_offset(selection)

        else:
            cells_wanted = dfc.multi_excel_ref_to_basecells(bounded)
            (
                self.min_selected_x,
                self.max_selected_x,
                self.min_selected_y,
                self.max_selected_y,
            ) = dfc.get_outlier_indicies(cells_wanted)

        # Where there's space, put a 1 cell blank border around walk
        # selections
        if selection_boundary:
            self.max_selected_x = (
                self.max_selected_x + 1
                if dfc.maximum_x_offset(selections[0].pcells) > self.max_selected_x
                else self.max_selected_x
            )
            self.min_selected_x = (
                self.min_selected_x - 1
                if dfc.minimum_x_offset(selections[0].pcells) < self.min_selected_x
                else self.min_selected_x
            )
            self.max_selected_y = (
                self.max_selected_y + 1
                if dfc.maximum_y_offset(selections[0].pcells) > self.max_selected_y
                else self.max_selected_y
            )
            self.min_selected_y = (
                self.min_selected_y - 1
                if dfc.minimum_y_offset(selections[0].pcells) < self.min_selected_y
                else self.min_selected_y
            )

    @property
    def highest_point(self):
        """
        The highest point of the table that we wish to preview, as you look at it
        """
        return self.min_selected_y

    @property
    def lowest_point(self):
        """
        The lowest point of the table that we wish to preview, as you look at it
        """
        return self.max_selected_y

    @property
    def leftmost_point(self):
        """
        The leftmost point of the table that we wish to preview, as you look at it
        """
        return self.min_selected_x

    @property
    def rightmost_point(self):
        """
        The rightmost point of the table that we wish to preview, as you look at it
        """
        return self.max_selected_x

    def contains(self, cell: Cell) -> bool:
        """
        Is the provided cell within the defined boundary of what we
        want to display as a preview.
        """
        return all(
            [
                cell.x >= self.min_selected_x,
                cell.x <= self.max_selected_x,
                cell.y >= self.min_selected_y,
                cell.y <= self.max_selected_y,
            ]
        )



from dataclasses import dataclass

from tidychef.against.implementations.base import BaseValidator
from tidychef.models.source.cell import Cell


class IsNumericOrFloatValidator(BaseValidator):
    """
    A class to return bool (valid or not) when
    called with a single instance of a tidychef
    Cell object.
    """

    def __call__(self, cell: Cell) -> bool:
        """
        Is the value property of the Cell
        numeric to include floats

        :param cell: A single tidychef Cell object.
        :return: bool, is it valid or not
        """
        if len(cell.value) > 0:
            if cell.value[0] == ".":
                return False
            if cell.value[-1] == ".":
                return False
        return cell.value.replace(".", "").isnumeric()

    def msg(self, cell: Cell) -> str:
        """
        Provide a contextually meaningful
        message to the user where cell
        is not numeric to include floats

        :param cell: A single tidychef Cell object.
        :return: A contextual message
        """

        return f'"{cell.value}" is not numeric or a float'


class IsNotNumericOrFloatValidator(BaseValidator):
    """
    A class to return bool (valid or not) when
    called with a single instance of a tidychef
    Cell object.
    """

    def __call__(self, cell: Cell) -> bool:
        """
        Is the value property of the Cell
        not numeric to include floats

        :param cell: A single tidychef Cell object.
        :return: bool, is it valid or not
        """
        if len(cell.value) > 0:
            if cell.value[0] == ".":
                return True
            if cell.value[-1] == ".":
                return True
        return not cell.value.replace(".", "").isnumeric()

    def msg(self, cell: Cell) -> str:
        """
        Provide a contextually meaningful
        message to the user where cell
        is not numeric to include floats

        :param cell: A single tidychef Cell object.
        :return: A contextual message
        """

        return f'"{cell.value}" is numeric or a float'


class IsNotNumericValidator(BaseValidator):
    """
    A class to return bool (valid or not) when
    called with a single instance of a tidychef
    Cell object.
    """

    def __call__(self, cell: Cell) -> bool:
        """
        Is the value property of the Cell
        not numeric

        :param cell: A single tidychef Cell object.
        :return: bool, is it valid or not
        """
        return not cell.value.isnumeric()

    def msg(self, cell: Cell) -> str:
        """
        Provide a contextually meaningful
        message to the user where cell
        is numeric

        :param cell: A single tidychef Cell object.
        :return: A contextual message
        """

        return f'"{cell.value}" is numeric'


class IsNumericValidator(BaseValidator):
    """
    A class to return bool (valid or not) when
    called with a single instance of a tidychef
    Cell object.
    """

    def __call__(self, cell: Cell) -> bool:
        """
        Is the value property of the Cell
        numeric

        :param cell: A single tidychef Cell object.
        :return: bool, is it valid or not
        """
        return cell.value.isnumeric()

    def msg(self, cell: Cell) -> str:
        """
        Provide a contextually meaningful
        message to the user where cell
        is numeric

        :param cell: A single tidychef Cell object.
        :return: A contextual message
        """

        return f'"{cell.value}" is not numeric'



"""
Holds the code that defines the local xlsx reader.
"""

from pathlib import Path
from typing import Callable, List, Optional, Union

import openpyxl

from tidychef.acquire.base import BaseReader
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import Table
from tidychef.selection.selectable import Selectable
from tidychef.selection.xlsx.xlsx import XlsxSelectable
from tidychef.utils import fileutils

from ..base import BaseReader
from ..main import acquirer


def local(
    source: Union[str, Path],
    selectable: Selectable = XlsxSelectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    tables: str = None,
    **kwargs
) -> Union[XlsxSelectable, List[XlsxSelectable]]:
    """
    Read data from a Path (or string representing a path)
    present on the same machine where tidychef is running.

    This local xlsx reader uses openpyxl:
    https://openpyxl.readthedocs.io/en/stable/index.html

    Additional kwargs passed to this function are propagated to
    the openpyxl.load_workbook() method.

    :param source: A Path object or a string representing a path
    :param selectable: A class that implements tidychef.selection.selectable.Selectable of an inheritor of. Default is XlsxSelectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of XlsxSelectable.parse() as an argument.
    :return: A single populated Selectable of type as specified by selectable param
    """

    assert isinstance(
        source, (str, Path)
    ), """
        The source you're passing to acquire.csv.local() needs to
        be either a Path object or a string representing such.
        """

    return acquirer(
        source,
        LocalXlsxReader(tables),
        selectable,
        pre_hook=pre_hook,
        post_hook=post_hook,
        **kwargs
    )


class LocalXlsxReader(BaseReader):
    """
    A reader to lead in a source where that source is a locally
    held xls file.
    """

    def parse(
        self,
        source: Union[str, Path],
        selectable: Selectable = XlsxSelectable,
        data_only: bool = True,
        **kwargs
    ) -> List[XlsxSelectable]:
        """
        Parse the provided source into a list of Selectables. Unless overridden the
        selectable is of type XlsxSelectable.

        Additional **kwargs are propagated to openpyxl.load_workbook()

        :param source: A Path or str representing a path indicating a local file
        :param selectable: The selectable type to be returned.
        :data_only: An openpyxl.load_workbook() option to disable acquisition of non data elements from the tabulated source (macros etc)
        :return: A list of type as specified by param selectable.
        """

        source: Path = fileutils.ensure_existing_path(source, **kwargs)

        workbook: openpyxl.Workbook = openpyxl.load_workbook(
            source, data_only=data_only, **kwargs
        )

        tidychef_selectables = []
        worksheet_names = workbook.get_sheet_names()
        for worksheet_name in worksheet_names:

            worksheet = workbook.get_sheet_by_name(worksheet_name)

            table = Table()
            for y, row in enumerate(worksheet.iter_rows()):
                for x, cell in enumerate(row):
                    table.add_cell(
                        Cell(x=x, y=y, value=str(cell.value) if cell.value else "")
                    )

            tidychef_selectables.append(
                selectable(table, source=source, name=worksheet_name)
            )
        return tidychef_selectables



import json
from dataclasses import dataclass
from typing import Dict, Optional

from tidychef.direction.directions import Direction
from tidychef.exceptions import AmbiguousLookupError, ImpossibleLookupError
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import LiveTable

from ..base import BaseLookupEngine

HIGHEST = 9999999999999
LOWEST = 0


@dataclass
class CellRange:
    """
    A class representing a range of cells

    :param low: the lowest index in the range
    :param high: the highest index in the range
    :param cell: the cell the spanning range resolves to
    :direction: the direction of traversal for the lookup
    """

    low: int
    high: int
    cell: Cell
    direction: Direction

    @property
    def axis_text(self) -> str:
        return "horizontal/x" if self.direction.is_horizontal else "vertical/y"

    def contains(self, cell: Cell) -> bool:
        """
        Does this cell contain the range
        in question.

        :param cell: The cell we want to check is in the range
        """
        if self.direction.is_horizontal:
            offset = cell.x
        else:
            offset = cell.y

        return offset >= self.low and offset <= self.high

    def spans_higher_range_than(self, cell: Cell) -> bool:
        """
        Does this range cover a range of indexes higher that the
        index of a given cell.

        :param cell: The cell this range may or may not span a higher
        range than.
        """
        if self.direction.is_horizontal:
            return self.low > cell.x
        else:
            return self.low > cell.y

    def spans_lower_range_than(self, cell: Cell) -> bool:
        """
        Does this range cover a range of indexes lower that the
        index of a given cell.

        :param cell: The cell this range may or may not span a lower
        range than.
        """
        if self.direction.is_horizontal:
            return self.high < cell.x
        else:
            return self.high < cell.y

    def _as_dict(self):
        return {
            "starts_at": str(self.low),
            "ends_at": str(self.high),
            "cell": str(self.cell),
            "axis": str(self.axis_text),
        }


class CellRanges:
    """
    A class representing multiple cell ranges
    """

    def __init__(self, selection: LiveTable, direction: Direction, table: str):
        """
        selection: The selection of cells to create the cell ranges from
        :direction: The direction of traversal considered when constructing ranges
        """
        self.direction: Direction = direction
        self.highest_possible_offset: Optional[int] = None
        self.lowest_possible_offset: Optional[int] = None
        self.table = table
        self._populate(selection)

    @property
    def axis_text(self) -> str:
        return "horizontal/x" if self.direction.is_horizontal else "vertical/y"

    def get_range_by_index(self, index: int) -> CellRange:
        """
        Get a single constructed range by index

        :param index: Index of the range to get
        """
        return self.ordered_cell_ranges[index]

    def _populate(self, selection: LiveTable):
        """
        Create the cell ranges from the selection provided

        :param selection: The selection of cells.
        """

        break_points = {}
        for cell in selection.cells:
            axis_offset = cell.x if self.direction.is_horizontal else cell.y
            if axis_offset in break_points.keys():
                raise AmbiguousLookupError(
                    f"""
                    Aborted while processing table "{self.table}.
                    
                    You have defined two or more equally valid closest:{self.direction.name}" relationships,
                    you cannot do this as it creates an ambiguous lookup.
                    
                    You are trying to add '{cell}' but we already have: '{break_points[axis_offset]}'.
                    
                    Both of these cells have a {self.axis_text} offsets of '{axis_offset}'. 
                    and so are equally close along this axis from any given observation cell.
                    """
                )
            break_points.update({axis_offset: cell})

        ordered_break_points = [int(k) for k in break_points.keys()]
        assert len(ordered_break_points) == len(selection.cells)
        ordered_break_points.sort()

        self.ordered_cell_ranges: Dict[int, CellRange] = {}

        if self.direction.is_left or self.direction.is_upwards:
            for i in range(0, len(ordered_break_points)):

                low = ordered_break_points[i]

                if i == 0:
                    self.lowest_possible_offset = low

                if i == len(ordered_break_points) - 1:
                    high = HIGHEST
                    self.highest_possible_offset = HIGHEST
                else:
                    high = ordered_break_points[i + 1] - 1

                self.ordered_cell_ranges[i] = CellRange(
                    low=low,
                    high=high,
                    cell=break_points[ordered_break_points[i]],
                    direction=self.direction,
                )

        else:
            for i in range(0, len(ordered_break_points)):

                high = ordered_break_points[i]

                if i == len(ordered_break_points) - 1:
                    self.highest_possible_offset = high

                if i == 0:
                    low = LOWEST
                    self.lowest_possible_offset = LOWEST
                else:
                    low = ordered_break_points[i - 1] + 1

                self.ordered_cell_ranges[i] = CellRange(
                    low=low,
                    high=high,
                    cell=break_points[ordered_break_points[i]],
                    direction=self.direction,
                )

        assert len(self.ordered_cell_ranges) == len(selection.cells)

    def _as_dict(self):
        """
        Reproduce the defined ranges as a dict.
        This functionality is purely for debugging and test purposes.
        """
        d = {}
        cell_range: CellRange
        for i, cell_range in self.ordered_cell_ranges.items():
            d.update({str(i): cell_range._as_dict()})
        return d


class Closest(BaseLookupEngine):
    def __init__(
        self,
        label: str,
        selection: LiveTable,
        direction: Direction,
        table: str = "Unnamed Table",
    ):
        """
        Creates a lookup engine to column values defined via
        the closest visual relationship.

        :param label: The label of the column informed
        by this lookup engine.
        :param: direction: one of up,down,left,right,above,below
        :param: Selection: the selection of cells that hold the column values being looked to.
        :param table: the name of the table data is being extracted from
        """
        self.table = table
        self.direction = direction
        self.ranges = CellRanges(selection, direction, table)
        self.bumped = True
        self.start_index = None
        self.label = label

    def _resolve_at_lower_range(self, index, cell, ceiling, floor):
        """
        Move the index down as as the cell offset was down/less-than the
        last ranged we looked at then attempt to resolve again.

        :param index: The index of the last range we looked at
        :param cell: The cell we're trying to find the correct range for
        :param ceiling: The highest range index to consider
        :param floor: The lowest range index to consider
        """

        if self.bumped == False and index != 0:
            new_index = index - 1
            self.bumped = True
        else:
            potential_range = ceiling - floor
            new_index = index - int(potential_range / 2)
            new_index = new_index if new_index != index else new_index - 1

        if new_index < 0:
            new_index = 0

        return self.resolve(cell, index=new_index, ceiling=index, floor=floor)

    def _resolve_at_higher_range(self, index, cell, ceiling, floor):
        """
        Move the index down as as the cell offset was above/greater-than the
        last ranged we looked at then attempt to resolve again.

        :param index: The index of the last range we looked at
        :param cell: The cell we're trying to find the correct range for
        :param ceiling: The lowest range index to consider
        :param floor: The highest range index to consider
        """

        if self.bumped == False and index != len(self.ranges.ordered_cell_ranges):
            new_index = index + 1
            self.bumped = True
        else:
            potential_range = ceiling - floor
            new_index = index + int(potential_range / 2)
            new_index = new_index if new_index != index else new_index + 1

        return self.resolve(cell, index=new_index, ceiling=ceiling, floor=index)

    def _confirm_within_bounds(self, cell: Cell):
        """
        Raise an exception if we're trying to resolve a lookup for a
        cell that is out of bounds, making a lookup impossible.

        example: we're trying to resolve a closest relations to the left
        of a cell that is further to the left than any of the ranges we're
        considering.

        :param cell: The single Cell object we want to resolve a the relative
        column cell object for.
        """

        err_str = """
                When processing table {table}.

                Lookup for observation cell '{cell}' is impossible. No
                cells in your selection exist in direction '{direction.name}'
                relative to this cell.

                The boundary (furthest cell index) out of the cells you provided
                on the `{axis}` axis has an offset of `{boundary}`.
                """

        out_of_bounds = [
            self.direction.is_upwards
            and cell.is_above(self.ranges.lowest_possible_offset),
            self.direction.is_downwards
            and cell.is_below(self.ranges.highest_possible_offset),
            self.direction.is_left
            and cell.is_left_of(self.ranges.lowest_possible_offset),
            self.direction.is_right
            and cell.is_right_of(self.ranges.highest_possible_offset),
        ]

        if True in out_of_bounds:
            axis = "x" if self.direction.is_horizontal else "y"
            raise ImpossibleLookupError(
                err_str.format(
                    table=self.table,
                    cell=cell,
                    direction=self.direction,
                    axis=axis,
                    boundary=self.ranges.lowest_possible_offset
                    if any([self.direction.is_upwards, self.direction.is_left])
                    else self.ranges.highest_possible_offset,
                )
            )

    def resolve(self, cell: Cell, index=None, ceiling=None, floor=0) -> Cell:
        """
        Given the cell we want to lookup the relative column value for, use
        a bisection search to identify the correct range in our ordered list of ranges.

        Note - this method is called recursively, using the kwargs to start
        again at a different point in the indexed list of ranges.

        :param index: The index of the last range we looked at
        :param cell: The cell we're trying to find the correct range for
        :param ceiling: The lowest range index to consider
        :param floor: The highest range index to consider
        :return: The tidychef Cell object representing the
        column cell.
        """

        # If neither the index or ceiling have been passed in, then its the
        # very first lookup - confirm cell is within the bounds and therefore
        # this lookup is theoretically possible.
        if not ceiling and not index:
            self._confirm_within_bounds(cell)

        # If no maximum cell range (ceiling) is set then its the absolute maximum
        if ceiling is None:
            ceiling: int = len(self.ranges.ordered_cell_ranges) - 1

        if index is None:
            # When starting range index is not passed in, start at the mid point.
            mid_point: int = int(len(self.ranges.ordered_cell_ranges) / 2)
            index = list(self.ranges.ordered_cell_ranges.keys())[mid_point]

        considered_range: CellRange = self.ranges.get_range_by_index(index)

        if considered_range.spans_lower_range_than(cell):
            return self._resolve_at_higher_range(
                index, cell, ceiling=ceiling, floor=floor
            )
        elif considered_range.spans_higher_range_than(cell):
            return self._resolve_at_lower_range(
                index, cell, ceiling=ceiling, floor=floor
            )
        else:
            # Lookup Caching
            # --------------
            # cells are implicitly selected right->down-a-row->right as you look at a tabulated
            # (i.e as per standard human reading convention) so we'll cache this range index as
            # there's a significant chance the next obs lookup is in the same range
            #
            # this right-then-down-then-right pattern also means that (often, not guaranteed),
            # if the next obs isn't in the last range checked, there's a decent chance it's in a
            # neighbouring range, so we'll also always "bump" the index once in the relevant
            # direction on a first miss.
            self.index = index
            self.bumped = False

            return considered_range.cell



class FileInputError(Exception):
    """
    There is an issues with what has been provided as a file input.
    """

    def __init__(self, msg: str):
        self.msg = msg


class CellsDoNotExistError(Exception):
    """
    User is trying to select something from the filtered table that
    does not exist in the filtered table.
    """

    def __init__(self, msg):
        self.msg = msg


class LoneValueOnMultipleCellsError(Exception):
    """
    Raised when a user attempts to use the Input.long_value() method
    on a selection of more than one cell.
    """

    def __init__(self, msg):
        self.msg = msg


class InvalidCellObjectError(Exception):
    """
    Raised where a cell object is missing required attributes or
    said attributes hold invalid types or values.
    """

    def __init__(self, msg):
        self.msg = msg


class UnalignedTableOperation(Exception):
    """
    Raised where a user it trying to combine cell selections from two
    distinctly separate inputs.

    This is invalid, as each input represents a subset (up to the whole of)
    a single distinct source of tabulated data.
    """

    def __init__(self, msg):
        self.msg = msg


class OutOfBoundsError(Exception):
    """
    Raised when a users attempts to select or reference a cell
    outside of the boundary of the pristine table of input cells.

    Example: Selecting every cell then shifting RIGHT 1 position
    would be trying to select a rightmost column of data that
    does not exist in the table.
    """

    def __init__(self, msg):
        self.msg = msg


class ZeroAcquiredTablesError(Exception):
    """
    Raised when a user is attempting to filter tables being
    acquired but it has resulted on no tables being acquired
    """

    def __init__(self, msg):
        self.msg = msg


class ImpossibleLookupError(Exception):
    """
    Raised when a user is attempting a lookup that is impossible
    to resolve.

    For example: looming up to a cell that is "above" directionally
    when no cells in the direction have been selected.
    """

    def __init__(self, msg):
        self.msg = msg


class MissingLabelError(Exception):
    """
    Raised where a user is trying to construct an output with a
    selection of cells but has not yet labelled that selection of
    cells
    """

    def __init__(self, msg):
        self.msg = msg


class BadConditionalResolverError(Exception):
    """
    Raised where a user has specified a HorizontalCondition
    lookup engine that has incorrect arguments or returns
    as incorrect types.
    """

    def __init__(self, msg):
        self.msg = msg


class HorizontalConditionalHeaderError(Exception):
    """
    Raised where a user has specified a HorizontalCondition
    lookup engine but its failing to resolve as an expected
    column value is missing.
    """

    def __init__(self, msg):
        self.msg = msg


class MisalignedHeadersError(Exception):
    """
    Raised where a user is attempting to join together two
    TidyData outputs but those outputs do not have the
    same column headers.
    """

    def __init__(self, msg):
        self.msg = msg


class DroppingNonColumnError(Exception):
    """
    Raised where a user is trying to drop a column
    during the transformation to tidy data but that
    column does not exist.
    """

    def __init__(self, msg):
        self.msg = msg



"""
A python package for easy and reliable extraction of messy data to create tidy data 
"""
from tidychef.notebook.preview.html.main import preview
from tidychef.selection import filters

from . import acquire, against, models, notebook, utils



"""
Common data functions that do not fall into any of the other categories.
"""
from typing import List, Tuple

from tidychef.exceptions import CellsDoNotExistError
from tidychef.models.source.cell import BaseCell


def assert_quadrilaterals(cells: List[BaseCell]) -> Tuple[int, int, int, int]:
    """
    Assert that the provided list of cells equates to selection of cells
    that form a quadrilateral shape with no gaps.

    This is a requirement when attempting to create an excel reference
    representing a list of cells.

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A tuple of ints, min_x, max_x, min_y, max_y.
    """

    min_x, max_x, min_y, max_y = get_outlier_indicies(cells)

    x_axis = (max_x - min_x) + 1
    y_axis = (max_y - min_y) + 1

    assert (x_axis * y_axis) == len(
        cells
    ), "The provided selection of cells does not equate to a quadrilateral selection"

    return min_x, max_x, min_y, max_y


def cell_is_within(cells: List[BaseCell], cell: BaseCell) -> bool:
    """
    Is the cell present within the list of cells

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :param cell: A BaseCell or inheritor of that represents a single selected cell.
    :return: bool, is the cell in the cells list.
    """
    find_cell = matching_xy_cells(cells, [cell])
    return len(find_cell) == 1


def cell_is_not_within(cells: List[BaseCell], cell: BaseCell) -> bool:
    """
    Is the cell absent from the list of cells

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :param cell: A BaseCell or inheritor of that represents a single selected cell.
    :return: bool, is the cell absent from the cells list.
    """
    return not cell_is_within(cells, cell)


def cells_not_in(
    initial_cells: List[BaseCell], unwanted_cells: List[BaseCell]
) -> List[BaseCell]:
    """
    Given two lists of cells. Return a List[BaseCell] representing
    initial_cells minus any cells from unwanted_cells

    :param initial_cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :param unwanted_cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: The cells from initial cells minus any that were in unwanted_cells
    """
    return [
        c1
        for c1 in initial_cells
        if not any(c1.matches_xy(c2) for c2 in unwanted_cells)
    ]


def cells_on_x_index(cells: List[BaseCell], x_index: int) -> List[BaseCell]:
    """
    Return a list from the provided cells that are on the specific x index.

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :xi: A horizontal index, the column number of a tabulated data source
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    return [c for c in cells if c.x == x_index]


def cells_on_y_index(cells: List[BaseCell], y_index: int) -> List[BaseCell]:
    """
    Return a list from the provided cells that are on the specific y index

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :param yi: A horizontal index, the column number of a tabulated data source
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    return [c for c in cells if c.y == y_index]


def exactly_matched_xy_cells(
    cells: List[BaseCell], wanted_cells: List[BaseCell]
) -> List[BaseCell]:
    """
    Given a list of cells, return any that match xy values
    with those in in wanted_cells.

    Raises an exception if we're asking for wanted_cells
    that do not exist.

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """

    unfound_cells = cells_not_in(wanted_cells, cells)

    if len(unfound_cells) > 0:
        raise CellsDoNotExistError(
            f"""
            Cannot find cell(s):
            {unfound_cells}

            In existing cells:
            {cells}
        """
        )

    return matching_xy_cells(cells, wanted_cells)


def exactly_matching_xy_cell(cells: List[BaseCell], wanted_cell: BaseCell) -> BaseCell:
    """
    Given a wanted cell, the cell from cells that matches xy values
    with it

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A single BaseCell or inheritor of.
    """
    match = [c1 for c1 in cells if any([c1.matches_xy(wanted_cell)])]
    assert len(match) == 1
    return match[0]


def get_outlier_indicies(cells: List[BaseCell]) -> Tuple[int, int, int, int]:
    """
    Given a list of cells, returns maximum and minimum x and y
    values from cells within that selection.

    returns:
    - min_x
    - max_x
    - min_y
    - max_y

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A tuple of ints, min_x, max_x, min_y, max_y.
    """
    min_x: int = minimum_x_offset(cells)
    max_x: int = maximum_x_offset(cells)
    min_y: int = minimum_y_offset(cells)
    max_y: int = maximum_y_offset(cells)
    return min_x, max_x, min_y, max_y


def matching_xy_cells(
    cells: List[BaseCell], wanted_cells: List[BaseCell]
) -> List[BaseCell]:
    """
    Given a list of cells, return all that match xy values
    with those in wanted_cells.

    Note: does NOT raise an exception if we're asking for wanted_cells
    that do not exist.

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    return [c1 for c1 in cells if any(c1.matches_xy(c2) for c2 in wanted_cells)]


def maximum_x_offset(cells: List[BaseCell]) -> int:
    """
    Given a list of BaseCell's, return the largest x position in use

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: The maximum horizontal index in use
    """
    max_x = max(c.x for c in cells)
    max_x_cell = [c for c in cells if c.x == max_x]
    return max_x_cell[0].x


def maximum_y_offset(cells: List[BaseCell]) -> int:
    """
    Given a list of BaseCell's, return the largest y position in use

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: The maximum vertical index in use.
    """
    max_y = max(c.y for c in cells)
    max_y_cell = [c for c in cells if c.y == max_y]
    return max_y_cell[0].y


def minimum_x_offset(cells: List[BaseCell]) -> int:
    """
    Given a list of BaseCell's, return the smallest x position in use

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: The minimum horizontal index in use.
    """
    min_x = min(c.x for c in cells)
    min_x_cell = [c for c in cells if c.x == min_x]
    return min_x_cell[0].x


def minimum_y_offset(cells: List[BaseCell]) -> int:
    """
    Given a list of BaseCell's, return the smallest y position in use

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: The maximum vertical index in use.
    """
    min_y = min(c.y for c in cells)
    min_y_cell = [c for c in cells if c.y == min_y]
    return min_y_cell[0].y


def specific_cell_from_xy(
    cells: List[BaseCell], x_index: int, y_index: int
) -> BaseCell:
    """
    Given a list of cells and specific x and y co-ordinates,
    return the requested cell.

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A single BaseCell or inheritor of.
    """
    cells_that_match = exactly_matched_xy_cells(cells, [BaseCell(x=x_index, y=y_index)])
    assert len(cells_that_match) == 1
    return cells_that_match[0]


def all_used_x_indicies(cells: List[BaseCell]) -> List[int]:
    """
    Given a list of cells, return each unique x indicies
    for a "row" that contains at least one cell.

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of every horizontal index (x attribute) present in cells.
    """
    return list(set((cell.x for cell in cells)))


def all_used_y_indicies(cells: List[BaseCell]) -> List[int]:
    """
    Given a list of cells, return each unique y indicies
    for a "column" that contains at least one cell.

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of every vertical index (y attribute) present in cells.
    """
    return list(set((cell.y for cell in cells)))



from typing import List

from .components import HtmlCell
from .constants import BORDER_CELL_COLOUR, INLINE_CSS, NO_COLOUR


def tidy_data_as_html_table_string(data: List[List[str]]) -> str:
    """
    When given tidy data in the form of a list of lists,i.e

    [
        ["row 1 value 1", "row 1 value 2"],
        ["row 2 value 1", "row 2 value 2"],
        etc...
    ]

    Return a html table representation of same.
    """

    headers = data[0]
    table_html = f"<tr>{''.join([HtmlCell(x, BORDER_CELL_COLOUR).as_html() for x in headers])}</tr>"

    rows = data[1:]
    for row in rows:
        table_html += (
            f"<tr>{''.join([HtmlCell(x, NO_COLOUR).as_html() for x in row])}</tr>"
        )

    return f"""
    <html>
        {INLINE_CSS}
            <body>
                <table>
                    {table_html}
                </table>
            </body>
            <br>
        </html>
    """



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Any, List, Optional, Union

from tidychef.selection.selectable import Selectable


@dataclass
class BaseReader(metaclass=ABCMeta):
    """
    Baseclass that all readers inherit from.
    """

    tables: Optional[str] = None

    @abstractmethod
    def parse(
        self, source: Any, selectable: Selectable, **kwargs
    ) -> Union[Selectable, List[Selectable]]:
        """Parse the datasource into a selectable thing"""



"""
Holds the code that defines the python list_of_lists reader.
"""

from os import linesep
from typing import Callable, List, Optional

from tidychef.acquire.base import BaseReader
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import Table
from tidychef.selection.selectable import Selectable

from ..base import BaseReader
from ..main import acquirer


def pipe_table(
    source: List[List[str]],
    selectable: Selectable = Selectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    **kwargs
) -> Selectable:
    """
    A reader to create a selectable from a string which represents a pipe
    delimited table. this is principally a convenience for testing.

    Example pipe table (without the escapes)

    = \"""
             |        |
        Male | Female | Both
        4    | 5      | 6
        7    | 8      | 9
      \"""

    :param source: A str representing a pipe table
    :param selectable: A class that implements tidychef.selection.selectable.Selectable of an inheritor of. Default is Selectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of ListOfListsReader.parse() as an argument.
    :return: A single populated Selectable of type as specified by selectable param
    """
    return acquirer(
        source,
        PipeTableReader(),
        selectable,
        pre_hook=pre_hook,
        post_hook=post_hook,
        **kwargs
    )


class PipeTableReader(BaseReader):
    def parse(self, source: str, selectable: Selectable = Selectable) -> Selectable:
        """
        Parse the provided source into a list of Selectables. Unless overridden the
        selectable is of type XlsSelectable.

        Additional **kwargs are propagated to xlrd.open_workbook()

        :param source: A list of lists of strings representing rows of cells
        :param selectable: The selectable type to be returned.
        :return: A populated instance of type as specified by param selectable.
        """
        table = Table()

        raw_rows = source.split(linesep)
        rows = []
        row_cell_count = None

        for row in raw_rows:
            if len(row.strip()) == 0:
                continue
            row_cells = [x.strip() for x in row.split("|")]
            if row_cell_count is None:
                row_cell_count = len(row_cells)
            else:
                assert (
                    len(row_cells) == row_cell_count
                ), "All rows in a pipe table must be the same length"
            rows.append(row_cells)

        for y_index, row in enumerate(rows):
            for x_index, cell_value in enumerate(row):
                table.add_cell(Cell(x=x_index, y=y_index, value=cell_value))

        return selectable(table)



"""
Holds the code that defines the local xlsx reader.
"""

import io
from pathlib import Path
from typing import Callable, List, Optional, Union

import requests
import validators
import xlrd

from tidychef.acquire.base import BaseReader
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import Table
from tidychef.selection.selectable import Selectable
from tidychef.selection.xls.xls import XlsSelectable
from tidychef.utils.http.caching import get_cached_session

from ..base import BaseReader
from ..main import acquirer


def http(
    source: Union[str, Path],
    selectable: Selectable = XlsSelectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    session: requests.Session = None,
    cache: bool = True,
    tables: str = None,
    **kwargs,
) -> Union[XlsSelectable, List[XlsSelectable]]:
    """
    Read data from a Path (or string representing a path)
    present on the same machine where tidychef is running.

    This xls reader uses xlrd:
    https://xlrd.readthedocs.io/en/latest/

    Any kwargs passed to this function are propagated to
    the xlrd.open_workbook() method.

    :param source: A url.
    :param selectable: A class that implements tidychef.selection.selectable.Selectable of an inheritor of. Default is XlsSelectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of HttpXlsReader.parse() as an argument.
    :param session: An optional requests.Session object.
    :param cache: Boolean flag for whether or not to cache get requests.
    :return: A single populated Selectable of type as specified by selectable param.
    """

    assert validators.url(source), f"'{source}' is not a valid http/https url."

    return acquirer(
        source,
        HttpXlsReader(tables),
        selectable,
        pre_hook=pre_hook,
        post_hook=post_hook,
        session=session,
        cache=cache,
        **kwargs,
    )


class HttpXlsReader(BaseReader):
    """
    A reader to lead in a source where that source is a locally
    held xlsx file.
    """

    def parse(
        self,
        source: str,
        selectable: Selectable = XlsSelectable,
        session: requests.Session = None,
        cache: bool = True,
        **kwargs,
    ) -> List[XlsSelectable]:
        """
        Parse the provided source into a list of Selectables. Unless overridden the
        selectable is of type XlsSelectable.

        Additional **kwargs are propagated to xlrd.open_workbook()

        :param source: A url
        :param selectable: The selectable type to be returned.
        :param session: An optional requests.Session object.
        :param session: An optional requests.Session object.
        :param cache: Boolean flag for whether or not to cache get requests.
        :return: A list of type as specified by param selectable.
        """

        if not session:
            if cache:
                session = get_cached_session()
            else:
                session = requests.session()

        response: requests.Response = session.get(source)
        if not response.ok:
            raise requests.exceptions.HTTPError(
                f"""
                Unable to get url: {source}
                {response}
                """
            )

        bio = io.BytesIO()
        bio.write(response.content)
        bio.seek(0)

        workbook: xlrd.Book = xlrd.open_workbook(file_contents=bio.read(), **kwargs)
        assert isinstance(workbook, xlrd.Book)

        tidychef_selectables = []
        worksheet_names = workbook.sheet_names()
        for worksheet_name in worksheet_names:

            worksheet = workbook.sheet_by_name(worksheet_name, **kwargs)

            table = Table()
            num_rows = worksheet.nrows
            for y in range(0, num_rows):
                for x, cell in enumerate(worksheet.row(y)):
                    table.add_cell(
                        Cell(x=x, y=y, value=str(cell.value) if cell.value else "")
                    )

            tidychef_selectables.append(
                selectable(table, source=source, name=worksheet_name)
            )
        return tidychef_selectables



import requests
from cachecontrol import CacheControl, serialize
from cachecontrol.caches.file_cache import FileCache
from cachecontrol.heuristics import LastModified


def get_cached_session() -> requests.Session:
    """
    Create a local requests cache in the hidden
    .cache directory.
    """
    session = CacheControl(
        requests.Session(),
        cache=FileCache(".cache"),
        serializer=serialize.Serializer(),
        heuristic=LastModified(),
    )
    return session



from abc import ABCMeta

from tidychef.lookup.base import BaseLookupEngine
from tidychef.models.source.cell import Cell


class BaseColumn(metaclass=ABCMeta):
    """
    A base class to hold the required variables for something
    that can be resolve to a column of data.
    """

    def __init__(
        self,
        engine: BaseLookupEngine,
        *args,
        **kwargs,
    ):
        """
        An object representing a proto column that can resolve a
        visual relationship to retrieve the appropriate column
        value relative to a given observation containing cell object.

        :param engine: The "lookup engine" that resolves cell value lookups.
        """

        # See _pre_init() docstring
        self._pre_init(self, engine, *args, **kwargs)

        assert isinstance(
            engine, BaseLookupEngine
        ), f"""
                {engine} is not a valid argument.
                
                Argument must be of type BaseLookupEngine.
        """

        self.engine = engine
        self.label = engine.label

        # See _post_init() docstring
        self._post_init(self, engine, *args, **kwargs)

    def _pre_init(
        self,
        engine: BaseLookupEngine,
        *args,
        **kwargs,
    ):
        """
        Overridable method for doing something clever just before the bulk
        of the __init__ logic is executed.

        :param engine: The "lookup engine" that resolves cell value lookups.
        """
        ...

    def _post_init(
        self,
        engine: BaseLookupEngine,
        *args,
        **kwargs,
    ):
        """
        Overridable method for doing something clever just after the __init__
        logic is executed.

        :param engine: The "lookup engine" that resolves cell value lookups.
        """
        ...

    def _post_lookup(self, cell: Cell):
        """
        Overridable method for doing something clever to a Cell that's just been
        resolved as belonging to this column.

        Example: prefix all cell value with "foo-"

        ```
            def _post_lookup(self, cell: Cell):
                cell.value = "foo-"+cell.value
                return Cell
        ``

        :param cell: A single tidychef Cell object.
        :return: A single tidychef Cell object.
        """
        return cell

    def resolve_column_cell_from_obs_cell(self, observation_cell: Cell, *args) -> Cell:
        """
        Use the provided lookup engine to return the value
        of this Column against a given observation, according
        to the lookup engine in use.

        The found cell is ran through _post_lookup() in case
        this particular Column is applying custom handling of
        some sort.

        :param cell: A single tidychef Cell object.
        :return: A single tidychef Cell object.
        """
        cell = self.engine.resolve(observation_cell, *args)
        cell = self._post_lookup(cell)
        return cell



import re
from dataclasses import dataclass
from re import Pattern
from typing import Optional

from tidychef.against.implementations.base import BaseValidator
from tidychef.models.source.cell import Cell


@dataclass
class RegexValidator(BaseValidator):
    pattern: str
    _compiled: Optional[Pattern] = None

    def __call__(self, cell: Cell) -> bool:
        """
        Does the value property of the Cell
        match the provided pattern.

        :param cell: A single tidychef Cell object.
        :return: bool, is it valid or not
        """
        if self._compiled is None:
            self._compiled = re.compile(r'' + self.pattern)
        if self._compiled.match(cell.value):
            return True
        return False

    def msg(self, cell: Cell) -> str:
        """
        Provide a contextually meaningful
        message to the user where cell
        value does not match the provided
        regular expression.

        :param cell: A single tidychef Cell object.
        :return: A contextual message
        """
        return f'"{cell.value}" does not match pattern: "{self.pattern}"'



import copy
from functools import wraps


def dontmutate(method):
    """
    Decorates a method so that any changes
    are made to a _copy_ of self not self.

    Without this, the following patterns:

    selection2 = selection1.do_something()

    match2 = match.regex("foo")

    Would change the value in the righthand
    classes (selection1 and match) as well
    as their lefthand assignations.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        self = copy.deepcopy(self)
        return method(self, *args, **kwargs)

    return wrapper



"""
Convenience wrappers for pre defined validators
"""

from typing import List, Optional

from .implementations.items import ItemsValidator
from .implementations.length import LengthValidator
from .implementations.numeric import (
    IsNotNumericOrFloatValidator,
    IsNotNumericValidator,
    IsNumericOrFloatValidator,
    IsNumericValidator,
)
from .implementations.regex import RegexValidator


def regex(pattern: str) -> RegexValidator:
    """
    Creates a RegexValidator to check if the
    tidychef Cell.value property of a given Cell
    matches the provided pattern.

    :param pattern: A regular expression
    :return: A instantiated RegexValidator
    """
    return RegexValidator(pattern)


def items(items: List[str]) -> ItemsValidator:
    """
    Creates an ItemsValidator to check if the
    tidychef Cell.value property of a given Cell
    is contained within the provided list.

    :param items: A list of strings representing valid values
    :return: A instantiated ItemsValidator
    """
    return ItemsValidator(items)


def length(least: Optional[int] = None, most: Optional[int] = None) -> LengthValidator:
    """
    Creates a LengthValidator to check if the
    tidychef Cell.value property of a given Cell
    is with the stated length constraints.

    :param least: The minimum length the cell value must be if any
    :param most: The minimum length the cell value must be if any
    :return: A instantiated LengthValidator
    """
    return LengthValidator(least=least, most=most)


# Pre instantiate these since no arguments are required.
is_numeric = IsNumericValidator()
is_not_numeric = IsNotNumericValidator()
is_not_numeric_or_float = IsNotNumericOrFloatValidator()
is_numeric_or_float = IsNumericOrFloatValidator()



"""
Holds the code that defines the local csv reader.
"""

import csv
import io
from typing import Any, Callable, Optional

import requests
import validators

from tidychef.acquire.base import BaseReader
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import Table
from tidychef.selection.csv.csv import CsvSelectable
from tidychef.selection.selectable import Selectable
from tidychef.utils.http.caching import get_cached_session

from ..base import BaseReader
from ..main import acquirer


def http(
    source: str,
    selectable: Selectable = CsvSelectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    session: requests.Session = None,
    cache: bool = True,
    **kwargs,
) -> CsvSelectable:
    """
    Creates a selectable with default child class CsvSelectable from
    a url with the http or https scheme.

    :param source: A url.
    :param selectable: A class that implements tidychef.selection.selectable.Selectable of an inheritor of. Default is CsvSelectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of HttpCsvReader.parse() as an argument.
    :param session: An optional requests.Session object.
    :param cache: Boolean flag for whether or not to cache get requests.
    :return: A single populated Selectable of type as specified by selectable param.
    """

    assert validators.url(source), f"'{source}' is not a valid http/https url."

    return acquirer(
        source,
        HttpCsvReader(),
        selectable,
        pre_hook=pre_hook,
        post_hook=post_hook,
        session=session,
        cache=cache,
        **kwargs,
    )


class HttpCsvReader(BaseReader):
    """
    A reader to read in a source where that source is a url
    representing a csv.
    """

    def parse(
        self,
        source: Any,
        selectable: Selectable = CsvSelectable,
        session: requests.Session = None,
        cache: bool = True,
        **kwargs,
    ) -> CsvSelectable:
        """
        Parse the provided source into a Selectable. Unless overridden the
        selectable is of type CsvSelectable.

        Optional **kwargs are propagated to the csv.reader() method.

        :param source: A url
        :param selectable: The selectable type to be returned.
        :param session: An optional requests.Session object.
        :param cache: Boolean flag for whether or not to cache get requests.
        :return: A list of type as specified by param selectable.
        """

        if not session:
            if cache:
                session = get_cached_session()
            else:
                session = requests.session()

        response: requests.Response = session.get(source)
        if not response.ok:
            raise requests.exceptions.HTTPError(
                f"""
                Unable to get url: {source}
                {response}
                """
            )

        sio = io.StringIO()
        sio.write(response.text)
        sio.seek(0)

        table = Table()
        file_content = csv.reader(sio, **kwargs)

        for y_index, row in enumerate(file_content):
            for x_index, cell_value in enumerate(row):
                table.add_cell(Cell(x=x_index, y=y_index, value=str(cell_value)))

        return selectable(table, source=source)



from abc import ABCMeta, abstractmethod


class BaseOutput(metaclass=ABCMeta):
    """
    A class to hold tidy data representations of a data source.
    """

    @abstractmethod
    def __str__(self):
        """
        What happens when someone prints this object.
        """

    @abstractmethod
    def __repr__(self):
        """
        What happens when someone wants to view a representation
        of this object.
        """

    @abstractmethod
    def _transform(self):
        """
        The logic for resolving the relationships.
        """



"""
Classes representing a sinlge cell of data.
"""
from __future__ import annotations

from dataclasses import dataclass
from os import linesep
from typing import Optional

from tidychef.exceptions import (
    InvalidCellObjectError,
    InvlaidCellPositionError,
    NonExistentCellComparissonError,
)
from tidychef.utils import cellutils

from .cellformat import CellFormatting


@dataclass
class BaseCell:
    """
    A primitive non value holding cell construct.

    :param x: Index on the horizontal axis
    :param y: Index on the vertical axis
    """

    x: int
    y: int

    def _confirm_not_virtual(self):
        """
        Confirms that a cell is not virtual and that
        the user us not attempting positional comparissons
        between an existent and non existent cell entity
        """
        if self.x is None or self.y is None:
            raise NonExistentCellComparissonError(
                "You cannot reference or use for comparison the "
                "positional information of a virtual cell as it "
                "does not exist in the source data, i.e it has no "
                f"position information.{linesep}"
                f'The value of the cell in question is "{self.value}"'
            )

    def matches_xy(self, other_cell: BaseCell) -> bool:
        """
        Does this objects x and y attributes, match
        the x and y attributes of the provided BaseCell or Cell
        object.

        :param other_cell: A different tidychef BaseCell or
        inheritor of.
        :return: True of cells have same table location, else False
        """
        self._confirm_not_virtual()
        return self.x == other_cell.x and self.y == other_cell.y

    def is_above(self, y: int) -> bool:
        """
        When compared to a y index, is this
        cell above it?

        We mean "above" in visual terms, i.e
        does it have a lower vertical offset
        from the top of the table.

        :param y: Vertical cell index (row number) to
        compare to.
        :return: is this Cell above this y
        """
        self._confirm_not_virtual()
        return self.y < y

    def is_below(self, y: int) -> bool:
        """
        When compared to a y index, is this
        cell below it?

        We mean "below" in visual terms, i.e
        does it have a higher vertical offset
        from the top of the table.

        :param y: Vertical cell index (row number) to
        compare to.
        :return: is this Cell below this y
        """
        self._confirm_not_virtual()
        return self.y > y

    def is_right_of(self, x: int) -> bool:
        """
        When compared to an x index, is this
        cell to the right of it?

        :param x: Horizontal cell index (column number) to
        compare to.
        :return: is this Cell right of x
        """
        self._confirm_not_virtual()
        return self.x > x

    def is_left_of(self, x: int) -> bool:
        """
        When compared to an x index, is this
        cell to the left of it?

        :param x: Horizontal cell index (column number) to
        compare to.
        :return: is this Cell left of x
        """
        self._confirm_not_virtual()
        return self.x < x

    def _excel_ref(self) -> str:
        """
        Get the excel reference of the cell

        :return: A string representation of this object.
        """

        x_ref = cellutils.x_to_letters(self.x) if self.x is not None else self.x
        y_ref = cellutils.y_to_number(self.y) if self.y is not None else self.y

        if any([x_ref and not y_ref, y_ref and not x_ref]):
            raise InvlaidCellPositionError(
                "Every cell object must have both an x and y position or neither."
                f"Got cell with x: {self.x} and y: {self.y}"
            )

        if x_ref and y_ref:
            return f"{x_ref}{y_ref}"
        return "VIRTUAL CELL"

    @property
    def excel_row(self) -> int:
        """
        What is the row number of the row
        containing this cell in excel terms
        """
        return cellutils.y_to_number(self.y)

    @property
    def excel_column(self) -> str:
        """
        What is the excel style column
        letter(s) for the column containing
        this cell.
        """
        return cellutils.x_to_letters(self.x)


@dataclass
class VirtualCell(BaseCell):
    """
    Where we are establishing relationships between a concrete cell
    from the tabulated data source and a constant or external value
    we do so via a virtual cell.

    VirtualCells are unique amongst cell variants in that they can
    have None positional values for x and y.
    """

    value: Optional[str] = None
    x: Optional[int] = None
    y: Optional[int] = None

    def __repr__(self):
        """
        Create a representation of this virtual cell in the form:
        <VIRTUAL CELL: value>
        """
        return f'({self._excel_ref()}, value:"{self.value}")'


@dataclass
class Cell(BaseCell):
    """
    Denotes a cell of data from a tabulated data source

    :param value: The contents of the cell as a string
    :param x: Index on the horizontal axis
    :param y: Index on the vertical axis
    :param cellformat: Optional attribute to hold format
    specific cell attributes
    """

    value: str
    x: int
    y: int

    # Optional as some tabulated formats (eg csv) do not have
    # cell formatting.
    cellformat: Optional[CellFormatting] = None

    def is_blank(self, disregard_whitespace: bool = True):
        """
        Can the contents of the cell be regarded as blank

        :param disregard_whitespace: Flag so we can choose to treat
        cells with just whitespace as populated.
        """
        if isinstance(self.value, str):
            v = self.value.strip() if disregard_whitespace else self.value
            if v == "":
                return True
            else:
                return False

        if not self.value:
            return True

        raise InvalidCellObjectError(
            f"Error with {self._as_xy_str()} A cell should have a str or nan/None value"
        )

    def is_not_blank(self, disregard_whitespace: bool = True):
        """
        Can the contents of the cell be regarded as not blank

        :param disregard_whitespace: Flag so we can choose to treat
        cells with just whitespace as populated.
        """
        return not self.is_blank(disregard_whitespace=disregard_whitespace)

    def _as_xy_str(self) -> str:
        """
        Returns a str representation of the current cell
        with xy co-ordinates and value.
        """
        return f'x:{self.x}, y:{self.y}, value = "{self.value}"'

    def __repr__(self):
        """
        Create a representation of this cell in the form:
        <excel ref: value>

        eg:
        <A1, value:"value of a1", x:{x}, y:{y}>
        """

        return f'({self._excel_ref()}, value:"{self.value}", x:{self.x}, y:{self.y})'



from pathlib import Path
from typing import Union

from tidychef.exceptions import FileInputError


def ensure_existing_path(maybe_path: Union[Path, str]) -> Path:
    """
    When given a Path or str representing an existing path, ensure that:

    a.) it is a Path (convert where necessary)
    b.) confirm it exists
    """

    if not isinstance(maybe_path, (Path, str)):
        raise FileInputError(
            "To use a direct file input, you must provide a pathlib.Path object or a str representing one"
        )

    if isinstance(maybe_path, str):
        maybe_path = Path(maybe_path)

    if not maybe_path.exists():
        raise FileInputError(
            f"A file at the path {maybe_path.absolute()} does not exist."
        )

    assert isinstance(maybe_path, Path)

    return maybe_path



"""
Holds the code that defines the python list_of_lists reader.
"""

from typing import Callable, List, Optional

from tidychef.acquire.base import BaseReader
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import Table
from tidychef.selection.selectable import Selectable

from ..base import BaseReader
from ..main import acquirer


def list_of_lists(
    source: List[List[str]],
    selectable: Selectable = Selectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    **kwargs
) -> Selectable:
    """
    A reader to create a selectable from a list of python
    lists, with each cell entry being a simple string.

    Regarding ordering we traverse the x axis then the y axis,
    i.e standard human reading order.

    For example:
    [
        ["Content of A1", "Contents of B1", "Contents of C1"],
        ["Content of A2", "Contents of B2", "Contents of C2"]
    ]

    :param source: A python list of lists
    :param selectable: A class that implements tidychef.selection.selectable.Selectable of an inheritor of. Default is Selectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of ListOfListsReader.parse() as an argument.
    :return: A single populated Selectable of type as specified by selectable param
    """
    return acquirer(
        source,
        ListOfListsReader(),
        selectable,
        pre_hook=pre_hook,
        post_hook=post_hook,
        **kwargs
    )


class ListOfListsReader(BaseReader):
    def parse(self, source, selectable: Selectable = Selectable) -> Selectable:
        """
        Parse the provided source into a list of Selectables. Unless overridden the
        selectable is of type XlsSelectable.

        Additional **kwargs are propagated to xlrd.open_workbook()

        :param source: A list of lists of strings representing rows of cells
        :param selectable: The selectable type to be returned.
        :return: A populated instance of type as specified by param selectable.
        """
        table = Table()

        assert (
            len(set([len(x) for x in source])) == 1
        ), "All rows must be the same length"

        for y_index, row in enumerate(source):
            for x_index, cell_value in enumerate(row):
                table.add_cell(Cell(x=x_index, y=y_index, value=str(cell_value)))

        return selectable(table)



"""
Classes representing a single cell of data.
"""

from __future__ import annotations

import copy
import uuid
from pathlib import Path
from typing import List, Optional, Union

import tidychef.datafuncs as dfc
from tidychef.exceptions import UnalignedTableOperation
from tidychef.models.source.cell import BaseCell, Cell
from tidychef.utils.decorators import dontmutate


class Table:
    """
    Represents a table of data in the form of a list of cell objects.
    """

    def __init__(self, cells: Optional[List[Cell]] = None):
        """
        Represents a table of data in the form of a list of cell objects.

        :param cells: A list of Cell objects representing the contents
        of a tabulated data source.
        """
        self.cells = None
        self._signature = str(uuid.uuid4())

    def add_cell(self, cell: Cell):
        if not self.cells:
            self.cells = []
        self.cells.append(cell)


class LiveTable:
    """
    A "live" table is the representation of a changing selection of cells as
    taken from a single tabulated input, it revolves around the control of
    two things:

    1.) "pristine" - The pristine table as pulled from the source.
    2.) "filtered" - The current subset of cells (up to all) as selected from the
                     pristine table.

    Keeping track of the pristine cell selection (the initial table) allows us to
    extend a Table of cells (.filtered) via comparing the two. This enables the
    easy extension of a cell selection as well as the filtering down of one.

    :param data_table: A tidychef Table object holding the cells representing
    a table.
    :param name: The name of the table where it has a name
    :param source: The filename, url or identifier of the source that has
    been ingested.
    """

    def __init__(self, data_table: Table, name: str = None, source: str = None):
        self.pristine: Table = data_table
        self.filtered: Table = copy.deepcopy(data_table)
        self._name: Optional[str] = name
        self.source: Optional[Union[Path, str]] = source

        # Label for a given selection
        self._label: Optional[str] = None

        # Cached
        self._maximum_pristine_x: Optional[int] = None
        self._minimum_pristine_x: Optional[int] = None
        self._maximum_pristine_y: Optional[int] = None
        self._minimum_pristine_y: Optional[int] = None

        # Config options
        self._explain: bool = False
        self._explain_path = None

    @property
    def maximum_pristine_x(self) -> int:
        """
        Get the minimum pristine y offset
        """
        if self._maximum_pristine_x is not None:
            return self._maximum_pristine_x
        else:
            self._maximum_pristine_x = dfc.maximum_x_offset(self.pcells)
            return self._maximum_pristine_x

    @property
    def minimum_pristine_x(self) -> int:
        """
        Get the minimum pristine y offset
        """
        if self._minimum_pristine_x is not None:
            return self._minimum_pristine_x
        else:
            self._minimum_pristine_x = dfc.minimum_x_offset(self.pcells)
            return self._minimum_pristine_x

    @property
    def maximum_pristine_y(self) -> int:
        """
        Get the minimum pristine y offset
        """
        if self._maximum_pristine_y is not None:
            return self._maximum_pristine_y
        else:
            self._maximum_pristine_y = dfc.maximum_y_offset(self.pcells)
            return self._maximum_pristine_y

    @property
    def minimum_pristine_y(self) -> int:
        """
        Get the minimum pristine y offset
        """
        if self._minimum_pristine_y is not None:
            return self._minimum_pristine_y
        else:
            self._minimum_pristine_y = dfc.minimum_y_offset(self.pcells)
            return self._minimum_pristine_y

    @property
    def label(self):
        """
        Accessor for current label of current selection of cells.
        """
        return self._label

    @dontmutate
    def label_as(self, label: str):
        """
        Assign a label to this specific selection

        :param label: The label we want to set.
        """
        assert isinstance(label, str), "A label for a selection must be of type string"
        self._label = label
        return self

    def __len__(self):
        return len(self.cells)

    @property
    def name(self):
        """
        Return a name if we have a name, else raises
        """
        if self._name:
            return self._name
        else:
            return "Unnamed Table"

    @property
    def cells(self) -> List[Cell]:
        """
        Accessor for currently selected cells from the
        currently selected table
        """
        return self.filtered.cells

    @cells.setter
    def cells(self, cells: List[Cell]):
        """
        Setter for the cells property

        :param cells: A lost of cells representing the
        currently selected cells from the table.
        """
        self.filtered.cells = cells

    @property
    def pcells(self) -> List[BaseCell]:
        """
        Accessor for the pristine cells from the
        currently selected table
        """
        return self.pristine.cells

    def selections_made(self) -> bool:
        """
        Have any selections been made
        """
        return len(self.pristine.cells) != len(self.filtered.cells)

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def signature(self) -> str:
        """
        A uuid that uniquely identifies a parsed input source table
        """
        return self.filtered._signature

    @dontmutate
    def __sub__(self, other_input: LiveTable):
        """
        Implements "-" operator, subtraction

        Allows subtraction of one selection from the same distinct
        and currently selected table from another. Provided they
        are derived from the same initial BaseInput.

        :param other_input: Another instance of this class
        whose selected cells we wish to subtract from the
        selected cells of this instance.
        """

        if self.signature != other_input.signature:
            raise UnalignedTableOperation(
                "Selections can only be combined or previewed in combination "
                "if they are taken from the exact same table as taken from a single "
                "instance of a parsed input."
            )

        cells = self.cells if self.selections_made() else self.pcells
        self.cells = dfc.cells_not_in(cells, other_input.cells)
        return self

    # TODO - type hint this
    @dontmutate
    def __or__(self, other_input: LiveTable):
        """
        Implements "|" operator, union.

        Allows the union of one selection from the same distinct
        and currently selected table with another. Provided they
        are derived from the same initial BaseInput.

        :param other_input: Another instance of this class
        whose cells we want to Union | with the cells of
        this class.
        """
        if self.signature != other_input.signature:
            raise UnalignedTableOperation(
                "Selections can only be combined or previewed in combination "
                "if they are taken from the exact same table as taken from a single "
                "instance of a parsed input."
            )

        new_cells = dfc.cells_not_in(other_input.cells, self.cells)
        self.cells = self.cells + new_cells
        return self

    # TODO - type hint this
    def __iter__(self):
        """ """
        for cell in self.cells:
            yield cell

    def __str__(self):
        return str(self.cells)



from typing import List

from tidychef import datafuncs as dfc
from tidychef.direction.directions import Direction
from tidychef.exceptions import ImpossibleLookupError, WithinAxisDeclarationError
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import LiveTable

from ..base import BaseLookupEngine


class Within(BaseLookupEngine):
    def __init__(
        self,
        label: str,
        selection: LiveTable,
        direction: Direction,
        start: Direction,
        end: Direction,
        table: str = "Unnamed Table",
    ):
        """
        Creates a lookup engine to resolve a Within lookup.

        Imagine the following example:

        |  A   |  B   |  C   |  D   |  E   |  F   |  G   |
        |------|------|------|------|------|------|------|
        |      |      |      |      |      |      |      |
        |      | age1 |      |      |      | age2 |      |
        | ob1  | ob2  | ob3  |      | ob4  | ob5  | ob6  |

        The relationships between the age dimension (age1 and age2 in the example)
        is both LEFT and RIGHT, relative to a given observation.

        Consider the following code:
        Within(cells, up, from=left(1), to=right(1))

        This is specifying age is above the observation, but only within the range of
        one to the left through 1 to the right.

        So in that example, we're searching (from the observation) for an age dimension
        cell via the following pattern.

        |   A  |  B   |  C   |  D   |
        |------|------|------|------|
        |   7  |  8   |  9   |      |
        |   4  |  5   |  6   |      |
        |   1  |  2   |  3   |      |
        |      |  ob  |      |      |

        With the header cell we're "looking up" being the first of the cells we've labelled
        1-9 (in that order) that is a cell we've also defined via the selection parameter.

        :param label: The label of the column informed
        by this lookup engine.
        :param Selection: the selection of cells that hold the column values being looked to.
        :param direction: cardinal direction declaration.
        :param start: a direction with an altered positional offset to indicate beginning of
        cell range to scan.
        :param end: a direction with an altered positional offset to indicate end of
        cell range to scan.
        :param table: the name of the table data is being extracted from
        """
        self.label = label
        self.table = table

        # Don't allow incorrect construction
        invalid_combinations = [
            start.is_horizontal and end.is_vertical,
            start.is_vertical and end.is_horizontal,
        ]

        if True in invalid_combinations:
            raise WithinAxisDeclarationError(
                f"""
                An error was encountered when processing column
                "{self.label}" of table "{self.table}" using a 
                within lookup.

                A Within class can only be constructed using two offset
                directions along a single axis.
                                             
                So you CAN combine
                - left, right
                - above, below
                - up, down
                etc
            
                You CANNOT combine
                - left, up
                - down, right
                etc
                                             
                As we're no longer scanning along a single axis.
                                             
                You provided:
                - name: "{start.name}", class: {start}
                - name: "{end.name}", class: {end} 
            """
            )

        self.start: Direction = start
        self.end: Direction = end
        self.direction_of_travel: Direction = end
        self.direction = direction

        self.cells: List[Cell] = selection.cells

    def _order(self, cells: List[Cell]) -> List[Cell]:
        """
        Order cells appropriately based on how the lookup engine has been configured

        :param cells: A list of cells to be ordered.
        :return: The ordered list of cells.
        """

        # Right
        if any([self.direction_of_travel.is_right and self.direction.is_downwards]):
            return dfc.order_cells_leftright_topbottom(cells)
        elif any([self.direction_of_travel.is_right and self.direction.is_upwards]):
            return dfc.order_cells_leftright_bottomtop(cells)

        # Left
        elif any([self.direction_of_travel.is_left and self.direction.is_downwards]):
            return dfc.order_cells_rightleft_topbottom(cells)
        elif any([self.direction_of_travel.is_left and self.direction.is_upwards]):
            return dfc.order_cells_rightleft_bottomtop(cells)

        # Downwards
        elif any([self.direction_of_travel.is_downwards and self.direction.is_right]):
            return dfc.order_cells_topbottom_leftright(cells)
        elif any([self.direction_of_travel.is_downwards and self.direction.is_left]):
            return dfc.order_cells_topbottom_rightleft(cells)

        # Upwards
        elif any([self.direction_of_travel.is_upwards and self.direction.is_right]):
            return dfc.order_cells_bottomtop_leftright(cells)
        elif any([self.direction_of_travel.is_upwards and self.direction.is_left]):
            return dfc.order_cells_bottomtop_rightleft(cells)

    def _feasible_cells(self, cell: Cell) -> List[Cell]:
        """
        Filters cells known to the engine to create a list of cells that are
        valid within the start= and end= params relative to the cell in question.

        :param cell: The tidychef cell we're trying to resolve the column
        cell for.
        :return: A list of cells that are feasible
        """

        if self.direction_of_travel.is_right and self.direction.is_upwards:
            x_start = cell.x + self.start.x
            x_end = cell.x + self.end.x
            return [
                c
                for c in self.cells
                if all([c.x >= x_start, c.x <= x_end, c.is_above(cell.y)])
            ]

        if self.direction_of_travel.is_left and self.direction.is_upwards:
            x_start = cell.x + self.end.x
            x_end = cell.x + self.start.x
            return [
                c
                for c in self.cells
                if all([c.x >= x_start, c.x <= x_end, c.is_above(cell.y)])
            ]

        if self.direction_of_travel.is_right and self.direction.is_downwards:
            x_start = cell.x + self.start.x
            x_end = cell.x + self.end.x
            return [
                c
                for c in self.cells
                if all([c.x >= x_start, c.x <= x_end, c.is_below(cell.y)])
            ]

        if self.direction_of_travel.is_left and self.direction.is_downwards:
            x_start = cell.x + self.end.x
            x_end = cell.x + self.start.x
            return [
                c
                for c in self.cells
                if all([c.x >= x_start, c.x <= x_end, c.is_below(cell.y)])
            ]

        if self.direction_of_travel.is_upwards and self.direction.is_right:
            y_start = cell.y + self.end.y
            y_end = cell.y + self.start.y
            return [
                c
                for c in self.cells
                if all([c.y >= y_start, c.y <= y_end, c.is_right_of(cell.x)])
            ]

        if self.direction_of_travel.is_upwards and self.direction.is_left:
            y_start = cell.y + self.end.y
            y_end = cell.y + self.start.y
            return [
                c
                for c in self.cells
                if all([c.y >= y_start, c.y <= y_end, c.is_left_of(cell.x)])
            ]

        if self.direction_of_travel.is_downwards and self.direction.is_left:
            y_start = cell.y + self.start.y
            y_end = cell.y + self.end.y
            return [
                c
                for c in self.cells
                if all([c.y >= y_start, c.y <= y_end, c.is_left_of(cell.x)])
            ]

        if self.direction_of_travel.is_downwards and self.direction.is_right:
            y_start = cell.y + self.start.y
            y_end = cell.y + self.end.y
            return [
                c
                for c in self.cells
                if all([c.y >= y_start, c.y <= y_end, c.is_right_of(cell.x)])
            ]

    def resolve(self, cell: Cell) -> Cell:
        """
        Given an observation cell, return the
        appropriate cell as declared via this
        visual relationship.

        :param cell: The tidychef Cell we want to resolve a column Cell for.
        :return: The column Cell we're resolved.
        """

        assert isinstance(cell, Cell)

        # Discard non feasible cells to avoid the overhead of ordering
        # cells that cant be the lookup
        feasible_cells = self._feasible_cells(cell)

        if len(feasible_cells) == 0:
            raise ImpossibleLookupError(
                f"""
                An error was encountered when processing column
                "{self.label}" of table "{self.table}" using a 
                within lookup.

                Within lookup cannot be resolved, there is no
                cell to find within the constraints:
                                        
                Start:  {self.start}
                End:    {self.end}

                Direction of traversal:
                {self.direction}

                Relative to observation cell {cell}

                Column cells were:
                {self.cells}                    
                """
            )
        ordered_feasible_cells = self._order(feasible_cells)
        assert len(ordered_feasible_cells) == len(feasible_cells)

        if self.direction.is_left:
            max_x = max([x.x for x in ordered_feasible_cells])
            for cell in ordered_feasible_cells:
                if cell.x == max_x:
                    return cell

        if self.direction.is_upwards:
            max_y = max([x.y for x in ordered_feasible_cells])
            for cell in ordered_feasible_cells:
                if cell.y == max_y:
                    return cell

        if self.direction.is_right:
            min_x = min([x.x for x in ordered_feasible_cells])
            for cell in ordered_feasible_cells:
                if cell.x == min_x:
                    return cell

        if self.direction.is_downwards:
            min_y = min([x.y for x in ordered_feasible_cells])
            for cell in ordered_feasible_cells:
                if cell.y == min_y:
                    return cell



from pathlib import Path
from typing import Dict, Union

from IPython.display import HTML, display

from tidychef.exceptions import OutputPassedToPreview, UnalignedTableOperation
from tidychef.models.source.table import LiveTable
from tidychef.output.base import BaseOutput

from .table import get_preview_table_as_html


def preview(
    *selections,
    path: Union[Path, str] = None,
    bounded: str = None,
    border_cells: str = "lightgrey",
    blank_cells: str = "white",
    warning_colour: str = "#ff8080",
    show_excel: bool = True,
    show_xy: bool = False,
    multiple_selection_warning: bool = True,
    selection_boundary: bool = False,
):
    """
    Create a preview from one of more selections of cells.

    :param *selections: 1-n tidychef selectables of inheritors of
    that will inform the preview.
    :param path: A Path or string representation of for where
    we want to write a html preview to local.
    :bounded param:
    """
    if isinstance(selections[0], BaseOutput):
        raise OutputPassedToPreview(
            """
            You cannot call preview an output.
            
            A preview displays the source data and the visual relationships you've
            you've declared within it. Once you have created an output (for example 
            created a TidyData() class) you can no longer preview it using this
            function.

            If you want to preview an output you can just print() it.
            """
        )

    for s in selections:
        assert isinstance(
            s, LiveTable
        ), f"Only selections and keyword arguments can be passed to preview, got {type(s)}"
    selections = list(selections)

    if len(set([s.signature for s in selections])) > 1:
        raise UnalignedTableOperation(
            "Selections can only be combined or previewed in combination "
            "if they are taken from the exact same table as taken from a single "
            "instance of a parsed input."
        )

    html_as_str = get_preview_table_as_html(
        selections,
        bounded=bounded,
        multiple_selection_warning=multiple_selection_warning,
        show_excel=show_excel,
        show_xy=show_xy,
        border_cells=border_cells,
        warning_colour=warning_colour,
        blank_cells=blank_cells,
        selection_boundary=selection_boundary,
    )

    if path:
        # If writing to an explain path, we write append
        write_mode = "a" if selection_boundary else "w"
        with open(path, write_mode) as f:
            f.write(html_as_str)
    else:
        display(HTML(html_as_str))



from typing import List

from tidychef import datafuncs as dfc
from tidychef.direction.directions import BaseDirection, Direction
from tidychef.exceptions import (
    FailedLookupError,
    MissingDirectLookupError,
    UnknownDirectionError,
)
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import LiveTable

from ..base import BaseLookupEngine


class Directly(BaseLookupEngine):
    """
    A class to resolve a direct lookup between
    a given observation cell and the appropriate
    cell from the selection of cells this class
    is constructed with.

    This appropriate cell is resolved based on the
    specified cardinal direction.

    Where no value exists in the direction specified
    an exception is raised.
    """

    def __init__(
        self,
        label: str,
        selection: LiveTable,
        direction: Direction,
        table: str = "Unnamed Table",
    ):
        """
        A class to resolve a direct lookup between
        a given observation cell and the appropriate
        cell from the selection of cells this class
        is constructed with.

        :param label: The label of the column informed
        by this lookup engine.
        :param: direction: one of up,down,left,right,above,below
        :param: Selection: the selection of cells that hold the column values being looked to.
        :param table: the name of the table data is being extracted from
        """
        self.table = table
        self.label = label
        self.direction: Direction = direction
        cells = selection.cells

        # Given we know the relationship is always
        # along a single axis, we'll create a
        # dict so we can just pluck out
        # the required lookup cell using the
        # relevant x or y offset of the
        # observation cell in question.

        # The complication comes from having
        # multiple potential selections along
        # a single axis.

        # example:
        # ob = observation
        # dim.* = dimension item we're looking up
        #
        # | dim1.1 |     | ob  | ob  |     | dim2.1 |     | [ob]  |  [ob] |
        # | dim1.2 |     | ob  | ob  |     | dim2.2 |     | [ob]  |  [ob] |
        # | dim1.3 |     | ob  | ob  |     | dim2.3 |     | [ob]  |  [ob] |
        # | dim1.4 |     | ob  | ob  |     | dim2.4 |     | [ob]  |  [ob] |
        # | dim1.5 |     | ob  | ob  |     | dim2.5 |     | [ob]  |  [ob] |
        #
        # If you consider each [ob] cells and a lookup for direction:left,
        # there are two available dimensions on that axis, we need to
        # differentiate the correct one per ob.

        if not isinstance(self.direction, BaseDirection):
            raise UnknownDirectionError(
                f"The direction parameter must be of type: {type(BaseDirection)}"
            )
        if self.direction.is_left or self.direction.is_upwards:
            ordered_cells = dfc.order_cells_leftright_topbottom(cells)
        elif self.direction.is_right or self.direction.is_downwards:
            ordered_cells = dfc.order_cells_rightleft_bottomtop(cells)
        else:
            # Shouldn't happen
            raise UnknownDirectionError(f"The direction {direction.name} is unknown.")

        self._lookups = {}
        for cell in ordered_cells:
            if self._index(cell) not in self._lookups:
                self._lookups[self._index(cell)] = []
            self._lookups[self._index(cell)].append(cell)

    def _index(self, cell: Cell):
        """
        Get the x or y offset we're interested in.

        By default its the along the principle direction
        of travel, i.e cell.x (column index) for a
        horizontal lookups else y (row index).

        :param cell: The tidychef Cell object we're trying
        to resolve the relative column cell object for.
        """
        if self.direction.is_horizontal:
            return cell.y
        return cell.x

    def resolve(self, cell: Cell) -> Cell:
        """
        Given an observation cell, return the
        appropriate cell as declared via this
        visual relationship.

        :param cell: The tidychef Cell object we're trying
        to resolve the relative column cell object for.
        :return: The tidychef Cell object representing the
        column cell.
        """

        potential_cells: List[Cell] = self._lookups.get(self._index(cell))
        if not potential_cells:
            raise MissingDirectLookupError(
                f"""
                When processing table "{self.table}" a direct lookup for column
                "{self.label}" failed because no column value exists in your
                column selection with direction: "{self.direction.name}" relative
                to the observation cell being resolved.
                
                The observation cell in question is:
                {cell._excel_ref()}, x position "{cell.x}", y position "{cell.y}",
                value: "{cell.value}"
            """
            )

        checker = {
            "left": lambda cell, pcell: cell.x > pcell.x,
            "right": lambda cell, pcell: cell.x < pcell.x,
            "up": lambda cell, pcell: cell.y > pcell.y,
            "above": lambda cell, pcell: cell.y > pcell.y,
            "down": lambda cell, pcell: cell.y < pcell.y,
            "below": lambda cell, pcell: cell.y < pcell.y,
        }

        chosen_cell = None
        for pcell in potential_cells:
            if checker[self.direction.name](cell, pcell):
                chosen_cell = pcell
            else:
                break

        if not chosen_cell:
            raise FailedLookupError(
                f"""
                When processing table "{self.table}" a Direct lookup for
                column "{self.label}" could not resolve with direction:
                "{self.direction.name}".

                The observation cell in question is:
                {cell._excel_ref()}, x position "{cell.x}", y position "{cell.y}",
                value: "{cell.value}"
                    """
            )

        return chosen_cell



from dataclasses import dataclass
from typing import Dict, List

from tidychef.models.source.cell import Cell
from tidychef.models.source.table import LiveTable

from .constants import COLOURS


class SelectionKey:
    """
    A coloured key denoting a single selection
    """

    def __init__(self, selection: LiveTable, colour_choice: int):
        """
        Creates a mapping between a colour choice
        and a named selection of cells.

        :param selection: A populated tidychef seletable
        or inheritor of.
        :param colour_choice: An integer specifying a colour
        from the constant list COLOURS
        """
        self.label = (
            selection._label
            if selection._label
            else f"Unnamed Selection: {colour_choice}"
        )
        self.colour = COLOURS[colour_choice]
        self.cells = selection.cells

    def matches_xy_of_cell(self, cell: Cell):
        """
        Is the provided tidychef Cell present in the
        selection of cells represented by this colour.

        :param cell: A single tidychef Cell object.
        """
        cells = [x for x in self.cells if x.x == cell.x and x.y == cell.y]
        return len(cells) == 1

    def as_html(self):
        """
        Create a html table entry identifying this key in previews.
        """
        return f"""
            <tr>
                <td style="background-color:{self.colour}">{self.label}</td>
            <tr>
        """


class SelectionKeys:
    """
    A holding class for constructing and iterating through multiple
    SelectionKey classes.
    """

    def __init__(self):
        self.colour_choice = 0
        self._keys: List[SelectionKey] = []

    def add_selection_key(self, selection: LiveTable):
        """
        Add a single SelectionKey to SelectionKeys

        :param selection: The single tidychef selectable or
        inheritor of that is represented by a single colour.
        """
        self._keys.append(SelectionKey(selection, self.colour_choice))
        self.colour_choice += 1

    def __iter__(self):
        """
        Iterates through the keys.
        """
        for selection_key in self._keys:
            yield selection_key


class HtmlCell:
    """
    Class to create a simple html representation of a single cell
    using a single background colour.
    """

    def __init__(self, value: str, colour: str):
        """
        Class to create a simple html representation of a single cell
        using a single background colour.

        :param value: The value contained in the cell in question.
        :param colour: The background colour to use for the cell.
        """
        self.value = value
        self.colour = colour

    def as_html(self):
        """
        Create the html representation of this cell.
        """
        return f'<td style="background-color:{self.colour}">{self.value}</td>'



"""
.. include:: ./README.md
"""
from .common import (
    all_used_x_indicies,
    all_used_y_indicies,
    assert_quadrilaterals,
    cell_is_not_within,
    cell_is_within,
    cells_not_in,
    cells_on_x_index,
    cells_on_y_index,
    exactly_matched_xy_cells,
    exactly_matching_xy_cell,
    get_outlier_indicies,
    matching_xy_cells,
    maximum_x_offset,
    maximum_y_offset,
    minimum_x_offset,
    minimum_y_offset,
    specific_cell_from_xy,
)
from .excel import (
    any_excel_ref_as_wanted_basecells,
    assert_excel_ref_within_cells,
    basecell_to_excel_ref,
    basecells_to_excel_ref,
    multi_excel_ref_to_basecells,
    single_excel_column_to_x_index,
    single_excel_ref_to_basecell,
    single_excel_row_to_y_index,
)
from .ordering import (
    order_cells_bottomtop_leftright,
    order_cells_bottomtop_rightleft,
    order_cells_leftright_bottomtop,
    order_cells_leftright_topbottom,
    order_cells_rightleft_bottomtop,
    order_cells_rightleft_topbottom,
    order_cells_topbottom_leftright,
    order_cells_topbottom_rightleft,
)



from IPython import get_ipython


# based on
# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
def in_notebook() -> bool:  # pragma: no cover
    """
    Checks whether the python interpreter calling
    this code is doing so via an IPython or Jupyter
    notebook.
    """
    if not get_ipython():
        return False
    if "IPKernelApp" not in get_ipython().config:
        return False
    return True



from __future__ import annotations

import copy
from typing import Callable, List, Dict, Optional

from tidychef.column.base import BaseColumn
from tidychef.exceptions import CellValidationError
from tidychef.lookup.base import BaseLookupEngine
from tidychef.lookup.engines.constant import Constant
from tidychef.lookup.engines.horizontal_condition import HorizontalCondition
from tidychef.models.source.cell import Cell


class Column(BaseColumn):
    """
    A basic implementation of a class that represents
    something that can be resolve into a column of data.

    This class differs from BaseColumn because:

    1. It allows the overriding of the value of cells that
    are extracted via the apply=keyword

    2. It allows the validation of data extracted via
    the validation= keyword.
    """

    _table: Optional[str] = "Unnamed Table"
    _apply_cache: Optional[Dict[Cell, Cell]] = None
    _validated_cells: Optional[List[Cell]] = None

    @property
    def table(self) -> str:
        """
        Returns table name where table name is known,
        else "Unnamed Table"
        """
        return self._table

    @staticmethod
    def horizontal_condition(column_label: str, resolver: Callable, priority=0) -> Column:
        """
        Creates a column that populates based on the
        values resolved for one or more other columns.

        :param column_label: The label we wish to give to the column.
        :param resolver: A callable to resolve the horizontal condition logic.
        :priority: controls order of resolution for all present HorizontalCondition objects,
        lower values are resolved first.
        :return: A Column object populated with a configured HorizontalCondition
        lookup engine.
        """
        return Column(HorizontalCondition(column_label, resolver, priority=priority))

    @staticmethod
    def constant(column_label: str, constant: str) -> Column:
        """
        Create a column that has one specific value for every
        entry.

        :param column_label: The label we wish to give to the column.
        :param constant: The constant value we wish to populate the column.
        :return: A Column object populated with a configured Constant lookup engine.
        """
        return Column(Constant(column_label, constant))

    def _pre_init(self, engine: BaseLookupEngine, *args, **kwargs):
        """
        Things to be applied before the bulk of the BaseColumn
        init logic.

        :engine: The lookup engine in use by this column.
        """

        # -----
        # Apply
        # -----
        self.apply = kwargs.get("apply", None)
        if self.apply:
            assert callable(
                self.apply
            ), "Value of Kwarg 'apply' must be a python callable"
            self._apply_cache = {}

        # ----------
        # Validation
        # ----------
        self.validation = kwargs.get("validate", None)
        if self.validation:
            assert callable(
                self.validation
            ), "Value of Kwarg 'validation' must be a python callable"
            self._validated_cells = []

    def _post_lookup(self, cell: Cell) -> Cell:
        """
        Makes use of apply and/or validation callables where the
        user has provided them.

        :param cell: A single instance of a tidychef Cell object.
        :return: A single instance of a tidychef Cell object.
        """

        # Apply
        # -----
        # We need to be careful not to change the Cell itself
        # as it might be called by other Columns.
        # Instead we:
        # (1) copy before modifying
        # (2) cache, so we only have to apply once per unique
        # input value (each column cell can resolve for many
        # observations)
        if self.apply is not None:
            already_applied_cell = self._apply_cache.get(cell.value, None)
            if already_applied_cell is None:
                applied_cell = copy.deepcopy(cell)
                applied_cell.value = self.apply(applied_cell.value)
                self._apply_cache[cell.value] = applied_cell
                cell = applied_cell
            else:
                cell = already_applied_cell

        # Validate
        # --------
        # Any _unique_ Cell value is either valid in this context or it
        # isn't, so only check its valid once.
        if self.validation is not None and cell not in self._validated_cells:
            self._validated_cells.append(cell)
            if not self.validation(cell):
                if hasattr(self.validation, "msg"):
                    msg = f"Message is: {self.validation.msg(cell)}"
                else:
                    msg = ''
                raise CellValidationError(f'''
                        Column {self.label} has a cell that is not valid,
                        Invalid cell {cell}.
                        {msg}                      
                        ''')

        return cell



class BadExcelReferenceError(Exception):
    """
    Raised where the user has provided an excel reference that does not match
    the known patterns for excel references.
    """

    def __init__(self, msg):
        self.msg = msg


class BadShiftParameterError(Exception):
    """
    Raised where someone has provided incorrect inputs to the
    shift method.
    """

    def __init__(self, msg):
        self.msg = msg


class CardinalDeclarationWithOffset(Exception):
    """
    User had tried to pass in an argument to a direction
    in a context where its just being used to declare an
    absolute cardinal direction.
    """

    def __init__(self, msg):
        self.msg = msg


class WithinAxisDeclarationError(Exception):
    """
    Raised where a user is attempting a within lookup but is
    mixing axis.

    i.e we must be scanning between two directions along a
    single axis, eg:

    - left to right
    - up to down
    - below to above
    etc

    We cannot scan up to left, above to right etc.
    """

    def __init__(self, msg):
        self.msg = msg


class OutputPassedToPreview(Exception):
    """
    User it trying to call preview() with a TidyData class
    in place of Selectables
    """

    def __init__(self, msg):
        self.msg = msg


class ReversedExcelRefError(Exception):
    """
    Raised where a user has provided an excel reference in a reversed format.

    Example:
    C5:A2
    """

    def __init__(self, msg):
        self.msg = msg


class UnknownDirectionError(Exception):
    """
    User has passed in a direction that is not a valid direction
    """

    def __init__(self, msg):
        self.msg = msg


class ReferenceOutsideSelectionError(Exception):
    """
    User is trying to make a narrow selection of cells
    but has already filtered the cells in question out
    of the table via a prior action.
    """

    def __init__(self, msg):
        self.msg = msg


class AmbiguousWaffleError(Exception):
    """
    User is trying to use the waffle method but has
    provided invalid combinations of selections.
    """

    def __init__(self, msg):
        self.msg = msg



"""
Source code for the acquirer class that power the data acquisition methods.

You would not typically be calling this directly outside of advanced users
utilising kwargs for unanticipated and/or niche uses cases.

In the vast majority of circumstances it is both easier and more reliable
to use the provided wrappers.

acquire.csv.local()
acquire.csv.remote()
etc...
"""
import re
from typing import Any, Callable, List, Optional, Union

from tidychef.acquire.base import BaseReader
from tidychef.exceptions import ZeroAcquiredTablesError
from tidychef.selection.selectable import Selectable


def acquirer(
    source: Any,
    reader: BaseReader,
    selectable: Selectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    **kwargs,
) -> Union[List[Selectable], Selectable]:
    """
    The principle data acquisition function. Wraps the reader
    to enable pre and post hook.

    :param source: A source appropriate for the provided BaseReader
    :param reader: A class that implements tidychef.acquire.base.BaseReader
    :param selectable: A class that implements tidychef.selection.selectable.Selectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of reader.parse() as an argument.
    :return: A single or list of class Selectable or inheritor of as returned by reader after
    optional modification by post_hook.
    """

    # Execute pre load hook
    if pre_hook:
        source = pre_hook(source)

    parsed = reader.parse(source, selectable, **kwargs)

    if reader.tables:
        assert isinstance(
            parsed, list
        ), "You can only use tables= where acquire is returning a list of selectabes."
        initial_table_names = [x.name for x in parsed]
        parsed = [x for x in parsed if re.match(reader.tables, x.name)]
        if len(parsed) == 1:
            parsed = parsed[0]
        if isinstance(parsed, list):
            if len(parsed) == 0:
                raise ZeroAcquiredTablesError(
                    f"""
                    The tables regex you provided: {reader.tables}
                    Has resulted in no tables being acquired.

                    The table names in question were:
                    {initial_table_names}
                    """
                )

    # Execute post load hook
    if post_hook:
        parsed = post_hook(parsed)

    return parsed



from abc import ABCMeta, abstractmethod

from tidychef.models.source.cell import Cell


class BaseValidator(metaclass=ABCMeta):
    """
    The standard Matcher used to validate a
    single cell object.
    """

    @abstractmethod
    def __call__(self, cell: Cell) -> bool:
        """
        Confirm that a single cell is valid.

        :param cell: A tidychef cell object
        :return: Is the cell in question valid.
        """

    @abstractmethod
    def msg(self, cell: Cell) -> str:
        """
        Generate a message on validation failure
        to provide some contextual information to
        the user
        """



import string


def letters_to_x(excel_letters_ref: str) -> int:
    """
    Given letters (as per excel column references), return the equivilent
    x co-ordinate
    """

    # Account for bad conventions
    excel_letters_ref = excel_letters_ref.upper()

    x = 0

    # Account for full iterations of the alphabet,
    # i.e AB, AAB, AH etc
    if len(excel_letters_ref) > 1:
        x = 26 * (len(excel_letters_ref) - 1)
        excel_letters_ref = excel_letters_ref[-1]

    for i, letter in enumerate(string.ascii_uppercase):
        if letter == excel_letters_ref:
            x += i
            break

    return x


def x_to_letters(x: int) -> str:
    """
    Convert an x co-ordinate to excel style letter references
    """

    # https://stackoverflow.com/questions/23861680/convert-spreadsheet-number-to-column-letter
    start_index = 0  #  it can start either at 0 or at 1
    letter = ""
    while x > 25 + start_index:
        letter += chr(65 + int((x - start_index) / 26) - 1)
        x = x - (int((x - start_index) / 26)) * 26
    letter += chr(65 - start_index + (int(x)))

    return letter


def number_to_y(excel_number_ref: int):
    """
    Given an excel row number, return the y offset
    """
    assert isinstance(excel_number_ref, int)
    return excel_number_ref - 1  # We are 0 indexed, unlike excel


def y_to_number(excel_number_ref: int):
    """
    Given a y offset, return the excel row number
    """
    assert isinstance(excel_number_ref, int)
    return excel_number_ref + 1  # We are 0 indexed, unlike excel



from __future__ import annotations

import copy
import os
import re
from os import linesep
from pathlib import Path
from typing import Callable, FrozenSet, List, Optional, Union

from tidychef import datafuncs as dfc
from tidychef.against.implementations.base import BaseValidator
from tidychef.direction.directions import (
    BaseDirection,
    Direction,
    above,
    below,
    down,
    left,
    right,
    up,
)
from tidychef.exceptions import (
    AmbiguousWaffleError,
    BadExcelReferenceError,
    BadShiftParameterError,
    CellsDoNotExistError,
    CellValidationError,
    LoneValueOnMultipleCellsError,
    MissingLabelError,
    OutOfBoundsError,
    ReferenceOutsideSelectionError,
)
from tidychef.lookup.engines.closest import Closest
from tidychef.lookup.engines.direct import Directly
from tidychef.lookup.engines.within import Within
from tidychef.models.source.cell import BaseCell, Cell
from tidychef.models.source.table import LiveTable
from tidychef.notebook.preview.html.main import preview
from tidychef.utils.decorators import dontmutate


class Selectable(LiveTable):
    """
    Inherits from LiveTable to add cell selection methods that are generic to all tabulated inputs.
    """

    def config(self, explain=False, explain_path: Optional[Union[str, Path]] = None):
        assert not all(
            [explain, explain_path]
        ), f"""
            Where you have specified explain_path= you do not need
            to also include explain=True. The keywords are mutually
            exclusive.
        """

        if explain_path is not None:
            if isinstance(explain_path, str):
                explain_path = Path(explain_path)
            if explain_path.exists():
                os.remove(explain_path.resolve())  # pragma: no cover

        self._explain_path = explain_path
        self._explain = explain

        return copy.deepcopy(self)

    def _get_excel_references(self) -> List[str]:
        """
        Returns a list of excel references for all selected cells in
        classical human readable order (left to right, top row to bottom row)
        """
        cells: List[Cell] = dfc.order_cells_leftright_topbottom(self.cells)
        return [x._excel_ref() for x in cells]

    def print_excel_refs(self):  # pragma: no cover
        """
        Orders cells in classical human readable order
        (right to left, top row to bottom row) then
        prints a list of ll excel references as a list
        """
        print(self._get_excel_references())

    def assert_len(self, number_of_cells: int):
        """
        Assert that the current selection contains the number of cells
        specified
        """
        assert (
            len(self.cells) == number_of_cells
        ), f"Selection contains {len(self.cells)} cells, not {number_of_cells}"
        _explain(self, f"WALK: Assert selection length is {number_of_cells}")
        return self

    def assert_one(self):
        """
        Assert that the current selection contains exactly one cell
        """
        return self.assert_len(1)

    def assert_single_column(self):
        """
        Assert that all CURRENTLY selected cells are contained in
        a single column.
        """
        assert (
            len(dfc.all_used_x_indicies(self.cells)) == 1
        ), "Selection has cells from more than one column"
        return self

    def assert_single_row(self):
        """
        Assert that all CURRENTLY selected cells are contained on
        a single row.
        """
        assert (
            len(dfc.all_used_y_indicies(self.cells)) == 1
        ), "Selection has cells from more than one row"
        return self

    def lone_value(self) -> str:
        """
        Confirms the selection contains exactly one cell, then returns
        the value of that cell
        """
        if len(self.cells) != 1:
            raise LoneValueOnMultipleCellsError(
                f"""
                Selection contains {len(self.cells)} cell. It must contain 1 to use .lone_value()
                """
            )
        return self.cells[0].value

    @dontmutate
    def is_blank(self, disregard_whitespace=True):
        """
        Filters the selection to those cells that are blank.
        By default a cell with just whitespace in it is
        considered blank. You can change this behaviour
        with the disregard_whitespace keyword.
        """
        self.cells = [
            x
            for x in self.cells
            if x.is_blank(disregard_whitespace=disregard_whitespace)
        ]
        _explain(self, f"Is blank")
        return self

    @dontmutate
    def is_not_blank(self, disregard_whitespace=True):
        """
        Filters the selection to those cells that are not blank.

        By default a cell with just whitespace in it is
        considered blank. You can change this behaviour
        with the disregard_whitespace keyword.
        """
        self.cells = [
            x
            for x in self.cells
            if x.is_not_blank(disregard_whitespace=disregard_whitespace)
        ]
        _explain(self, f"Is not blank")
        return self

    @dontmutate
    def expand(self, direction: Direction):
        """
        Given a direction of UP, DOWN, LEFT, RIGHT
        Expands the current selection of cells in that direction.

        Notes:
        - Will also accept ABOVE and BELOW as direction, as they
        are aliases of UP and DOWN respectively.
        """
        direction._confirm_pristine()
        selection: List[BaseCell] = []

        # To begin with, the potential cells is equal to all cells
        # not currently selected.
        potential_cells: List[Cell] = dfc.cells_not_in(self.pcells, self.cells)

        if direction in [up, down, above, below]:

            # Only consider things on the same horizontal(x) index as a cell
            # that's already selected.
            all_used_x_indicies: FrozenSet[int] = set(c.x for c in self.cells)

            # Now consider each relevant x index
            for xi in all_used_x_indicies:

                # All cells on this index (i.e in this column)
                selected_cells_on_xi = dfc.cells_on_x_index(self.cells, xi)

                # Not currently selected cells on this index
                potential_cells_on_xi: List[Cell] = [
                    c for c in potential_cells if c.x == xi
                ]

                if direction in [up, above]:
                    lowest_used_xi = dfc.maximum_y_offset(selected_cells_on_xi)
                    # Add cells from the potential selection to the
                    # actual selection if they meet the criteria.
                    selection += [
                        c
                        for c in potential_cells_on_xi
                        if c.is_above(lowest_used_xi)  # above: visually
                    ]

                if direction in [down, below]:
                    largest_used_yi = dfc.minimum_y_offset(selected_cells_on_xi)
                    # Add cells from the potential selection to the
                    # actual selection if they meet the criteria.
                    selection += [
                        c
                        for c in potential_cells_on_xi
                        if c.is_below(largest_used_yi)  # below: visually
                    ]

        if direction in [left, right]:

            # For every row in which have at least one cell selected
            all_used_y_indicies: FrozenSet[int] = set(c.y for c in self.cells)
            for yi in all_used_y_indicies:

                # Get all currently selected cells on that row
                selected_cells_on_yi = dfc.cells_on_y_index(self.cells, yi)

                # Get all not selected cells on that row
                potential_cells_on_yi: List[Cell] = [
                    c for c in potential_cells if c.y == yi
                ]

                if direction == left:

                    # Select anything to the left of the
                    # rightmost of the selected cells on this row
                    leftmost_used_yi = dfc.minimum_x_offset(selected_cells_on_yi)
                    selection += [
                        c
                        for c in potential_cells_on_yi
                        if c.is_left_of(leftmost_used_yi)
                    ]

                if direction == right:
                    rightmost_used_yi = dfc.maximum_x_offset(selected_cells_on_yi)
                    selection += [
                        c
                        for c in potential_cells_on_yi
                        if c.is_right_of(rightmost_used_yi)
                    ]

        self.cells += selection
        _explain(self, f"Expand: {direction.name}")
        return self

    @dontmutate
    def fill(self, direction: Direction):
        """
        Creates a new selection from the cells in that direction
        relative to the current cell selection.

        :direction: One of: up, down, left, right
        """

        # Fill is just a slightly modified wrapper
        # for expand. So if ._explain is on we need
        # to toggle if off while doing the expand
        # to avoid confusing the user
        explain_setting = self._explain
        explain_path_setting = self._explain_path
        self._explain = False
        self._explain_path = None

        did_have = copy.deepcopy(self.cells)
        self = self.expand(direction)

        self._explain = explain_setting
        self._explain_path = explain_path_setting

        self.cells = [x for x in self.cells if x not in did_have]
        _explain(self, f"Fill: {direction.name}")
        return self

    @dontmutate
    def shift(
        self,
        direction_or_x: Union[Direction, int],
        possibly_y: Optional[int] = None,
    ):
        """
        Move the entire current selection relatively, examples:

        - .shift(right)
        - .shift(right(5))
        - .shift(2, 6)
        - .shift(-1, 4)

        :param direction_or_x: Either a direction of the raw x offset
        of a direction.
        :param possibly_y: Either none or the raw y offset of a direction
        """

        msg = (
            (
                f"The shift method must be called with one of two types of argument{linesep}"
                f"1.) By passing in an up, down, left, right, above or below direction, "
                f"for example: .shift(up). {linesep}"
                "2.) By passing in two integer arguments, on each for x index change and y index change"
                "example: .shift(1, 2)"
            ),
        )

        if isinstance(direction_or_x, int):
            if not isinstance(possibly_y, int):
                raise BadShiftParameterError(msg)
            x_offset = direction_or_x
            y_offset = possibly_y
        elif isinstance(direction_or_x, BaseDirection):
            assert (
                not possibly_y
            ), "Where passing a direction into shift, that must be the only argument"
            x_offset = direction_or_x.x
            y_offset = direction_or_x.y
        else:
            raise BadShiftParameterError(msg)

        wanted_cells: List[BaseCell] = [
            BaseCell(x=c.x + x_offset, y=c.y + y_offset) for c in self.cells
        ]

        found_cells = dfc.matching_xy_cells(self.pcells, wanted_cells)

        if len(found_cells) == 0 and len(wanted_cells) > 0:
            raise OutOfBoundsError(
                "You are attempting to shift your selection "
                "entirely outside of the boundary of the table."
            )

        self.cells = found_cells

        if not possibly_y:
            _explain(self, f"Shifted cells {direction_or_x.offset_as_str}")
        else:
            _explain(self, f"Shifted cells {direction_or_x}, {possibly_y}")
        return self

    @dontmutate
    def excel_ref(self, excel_ref: str):
        """
        Selects just the cells as indicated by the provided excel style
        reference: "A6", "B17:B24", "9", "GH" etc.
        """

        msg = f"""
                You cannot make a selection of "{excel_ref}" at
                this time. One or more cells of these cells
                does not exist in your CURRENT SELECTION.
                
                If you believe they should be, you need to check
                your sequencing.

                As an example: if you filter a selection to just
                column A then try excel_ref('B') you'll get
                this error.

                Where practical, you can debug the selected cells
                at any given time with <selection>.print_excel_refs()
                """

        # Multi excel reference:
        # eg: 'B2:F5'
        if re.match("^[A-Z]+[0-9]+:[A-Z]+[0-9]+$", excel_ref):
            cell1 = dfc.single_excel_ref_to_basecell(excel_ref.split(":")[0])
            cell2 = dfc.single_excel_ref_to_basecell(excel_ref.split(":")[1])
            try:
                assert cell1.x >= self.minimum_pristine_x
                assert cell1.x <= self.maximum_pristine_x
                assert cell1.y >= self.minimum_pristine_y
                assert cell1.y <= self.maximum_pristine_y
                assert cell2.x >= self.minimum_pristine_x
                assert cell2.x <= self.maximum_pristine_x
                assert cell2.y >= self.minimum_pristine_y
                assert cell2.y <= self.maximum_pristine_y
                wanted: List[BaseCell] = dfc.multi_excel_ref_to_basecells(excel_ref)
                selected = dfc.exactly_matched_xy_cells(self.cells, wanted)
            except CellsDoNotExistError:
                raise ReferenceOutsideSelectionError(msg)
            except AssertionError:
                raise ReferenceOutsideSelectionError(msg)

        # Single column and row reference
        # eg: 'F19'
        elif re.match("^[A-Z]+[0-9]+$", excel_ref):
            wanted: BaseCell = dfc.single_excel_ref_to_basecell(excel_ref)
            wanted = [wanted]
            try:
                selected = dfc.exactly_matched_xy_cells(self.cells, wanted)
            except CellsDoNotExistError:
                raise ReferenceOutsideSelectionError(msg)

        # An excel reference that is a single row number
        # eg: '4'
        elif re.match("^[0-9]+$", excel_ref):
            wanted_y_index: int = dfc.single_excel_row_to_y_index(excel_ref)
            wanted = [c for c in self.pcells if c.y == wanted_y_index]
            try:
                assert wanted_y_index <= self.maximum_pristine_y
                assert wanted_y_index >= self.minimum_pristine_y
                selected = dfc.exactly_matched_xy_cells(self.cells, wanted)
            except CellsDoNotExistError:
                raise ReferenceOutsideSelectionError(msg)
            except AssertionError:
                raise ReferenceOutsideSelectionError(msg)

        # An excel reference that is a multiple row numbers
        # eg: '4:6'
        elif re.match("^[0-9]+:[0-9]+$", excel_ref):
            start_y = excel_ref.split(":")[0]
            end_y = excel_ref.split(":")[1]
            start_y_index: int = dfc.single_excel_row_to_y_index(start_y)
            end_y_index: int = dfc.single_excel_row_to_y_index(end_y)
            if start_y_index >= end_y_index:
                raise BadExcelReferenceError(
                    f'Excel ref "{excel_ref}" is invalid. {end_y_index} must be higher than {start_y_index}'
                )
            try:
                assert end_y_index <= self.maximum_pristine_y
                assert start_y_index >= self.minimum_pristine_y
                wanted = [
                    c
                    for c in self.pcells
                    if c.y >= start_y_index and c.y <= end_y_index
                ]
                selected = dfc.exactly_matched_xy_cells(self.cells, wanted)
            except CellsDoNotExistError:
                raise ReferenceOutsideSelectionError(msg)
            except AssertionError:
                raise ReferenceOutsideSelectionError(msg)

        # An excel reference that is one column letter
        # eg: 'H'
        elif re.match("^[A-Z]+$", excel_ref):
            wanted_x_index: int = dfc.single_excel_column_to_x_index(excel_ref)
            try:
                assert wanted_x_index <= self.maximum_pristine_x
                assert wanted_x_index >= self.minimum_pristine_x
                wanted = [c for c in self.pcells if c.x == wanted_x_index]
                selected = dfc.exactly_matched_xy_cells(self.cells, wanted)
            except CellsDoNotExistError:
                raise ReferenceOutsideSelectionError(msg)
            except AssertionError:
                raise ReferenceOutsideSelectionError(msg)

        # An excel reference that is a range of column letters
        # eg: 'H:J'
        elif re.match("^[A-Z]+:[A-Z]+$", excel_ref):
            left_letters = excel_ref.split(":")[0]
            right_letters = excel_ref.split(":")[1]
            start_x_index: int = dfc.single_excel_column_to_x_index(left_letters)
            end_x_index: int = dfc.single_excel_column_to_x_index(right_letters)
            if start_x_index >= end_x_index:
                raise BadExcelReferenceError(
                    f'Excel ref "{excel_ref}" is invalid. {right_letters} much be higher than {left_letters}'
                )
            wanted = [
                c for c in self.pcells if c.x >= start_x_index and c.x <= end_x_index
            ]
            try:
                assert end_x_index <= self.maximum_pristine_x
                assert start_x_index >= self.minimum_pristine_x
                selected = dfc.exactly_matched_xy_cells(self.cells, wanted)
            except CellsDoNotExistError:
                raise ReferenceOutsideSelectionError(msg)
            except AssertionError:
                raise ReferenceOutsideSelectionError(msg)

        # Unknown excel reference
        else:
            raise BadExcelReferenceError(f"Unrecognised excel reference: {excel_ref}")

        self.cells = selected
        _explain(self, f"Excel reference: {excel_ref}")
        return self

    def validate(self, validator: BaseValidator, raise_first_error: bool = False):
        """
        Validates current cell selection by passing each currently
        selected cell to the provided validator.

        Pass raise_first_error=True if you just want the first
        invalid value message.
        """

        validation_errors = []
        for cell in self.cells:
            if not validator(cell):
                if raise_first_error:
                    raise CellValidationError(f'''
When making selections from table: {self.name} the
following validation error was encountered:
{validator.msg(cell)}
                ''')
                else:
                    validation_errors.append(validator.msg(cell))

        if len(validation_errors) > 0:
            raise CellValidationError(f'''
When making selections from table: {self.name} the
following validation errors were encountered:
{linesep.join(validation_errors)}
                ''')

        return self

    @dontmutate
    def filter(self, check: Callable):
        """
        Selects just the cells that match the provided check

        : param check: a function, lambda or callable class that
        returns a bool when given a single cell as a parameter.
        """

        self.cells = list(filter(check, self.cells))
        if hasattr(check, "explain"):
            comment = check.explain
        else:
            comment = "Custom filter"
        _explain(self, f"Filtered: {comment}")
        return self

    @dontmutate
    def re(self, pattern: str):
        """
        Filter the current selection of cells to only include
        cells whose value matches the provided regular expression
        pattern.
        """

        matcher = re.compile(r'' + pattern)
        self.cells = [x for x in self.cells if matcher.match(x.value) is not None]
        _explain(self, f"Regex, pattern {pattern}")
        return self

    @dontmutate
    def waffle(self, direction: Direction, additional_selection: Selectable):
        """
        A "waffle" will select all cells that directionally intersect
        with both a cell in the current selection and a cell in the
        selection being passed into this method.

        Examples:

        [B1].waffle("A6") == [B6]

        [C4,C5,C6].waffle("F1", "G1") == [F4,F5,F6,G4,G5,G6]
        """

        if direction.is_vertical:
            if direction.is_downwards:
                highest_y = dfc.maximum_y_offset(self.cells)
                if any([x for x in additional_selection if x.y <= highest_y]):
                    raise AmbiguousWaffleError(
                        "When using waffle down, your additional selections must all "
                        "be below your initial selections."
                    )
            if direction.is_upwards:
                lowest_y = dfc.minimum_y_offset(self.cells)
                if any([x for x in additional_selection if x.y >= lowest_y]):
                    raise AmbiguousWaffleError(
                        "When using waffle up, your additional selections must all be "
                        "above your initial selections."
                    )
            x_offsets = dfc.all_used_x_indicies(self.cells)
            y_offsets = dfc.all_used_y_indicies(additional_selection.cells)
        else:
            if direction.is_right:
                highest_x = dfc.maximum_x_offset(self.cells)
                if any([x for x in additional_selection if x.x <= highest_x]):
                    raise AmbiguousWaffleError(
                        "When using waffle right, your additional selections must all "
                        "be right of your initial selections."
                    )
            if direction.is_left:
                lowest_x = dfc.minimum_x_offset(self.cells)
                if any([x for x in additional_selection if x.x >= lowest_x]):
                    raise AmbiguousWaffleError(
                        "When using waffle left, your additional selections must all be "
                        "left of your initial selections."
                    )
            x_offsets = dfc.all_used_x_indicies(additional_selection.cells)
            y_offsets = dfc.all_used_y_indicies(self.cells)

        self.cells = [x for x in self.pcells if x.x in x_offsets and x.y in y_offsets]

        return self

    @dontmutate
    def extrude(self, direction: Direction):
        """
        Increase selection in a single direction s by the amount
        passed as an argument to direction. Where no integer
        parameter is passed into direction, the size of the
        extrusion is one.

        Examples:

        .extrude(right(2)) - increase selection by all cells
        that are within two cells right of a currently selected
        cell.

        .extrude(down(3)) - increase selection by all cells
        that are within three cells down of a currently selected
        cell.

        .extrude(left) - increase selection by select all
        cells that are one cell left of a currently selected cell.

        :param direction: A direction of up, down, left, right,
        above below with optional offset parameter.
        """

        additional_cells = []
        for cell in self.cells:

            if direction.is_horizontal:
                viable_cells = dfc.cells_on_y_index(self.pcells, cell.y)
                if direction.is_right:
                    extruded_cells = [
                        x
                        for x in viable_cells
                        if x.x > cell.x and x.x <= (cell.x + direction.x)
                    ]
                    additional_cells + extruded_cells
                if direction.is_left:
                    extruded_cells = [
                        x
                        for x in viable_cells
                        if x.x < cell.x and x.x >= (cell.x + direction.x)
                    ]
                additional_cells += extruded_cells

            else:
                viable_cells = dfc.cells_on_x_index(self.pcells, cell.x)
                if direction.is_downwards:
                    extruded_cells = [
                        x
                        for x in viable_cells
                        if x.y > cell.y and x.y <= (cell.y + direction.y)
                    ]
                if direction.is_upwards:
                    extruded_cells = [
                        x
                        for x in viable_cells
                        if x.y < cell.y and x.y >= (cell.y + direction.y)
                    ]
                additional_cells += extruded_cells

        self.cells += [x for x in additional_cells if x not in self.cells]
        _explain(self, f"Extrude: {direction.offset_as_str}")
        return self

    def finds_observations_directly(self, direction: Direction) -> Directly:
        """
        Creates and returns a class:Directly lookup engine
        that can resolve the correct cell from this selection
        relative to any given observation.
        """

        if not self._label:
            raise MissingLabelError(
                """
                You are trying to create a lookup engine for a selection of
                cells using the .resolve_observations_directly() method but
                have not yet assigned a label to said selection of cells.

                Please use the .label_as() method to assign a label before
                attempting this.
            """
            )

        # The constructor we provide to the user advertises that the column
        # lookup engines "finds the observations" but this is purely a
        # conceptual trick to make a more user friendly api.
        # In reality that's exactly backwards, an observation actually finds
        # a column value (by being passed to the lookup engine constructed below).
        # As such we need to reverse the stated direction.
        return Directly(self.label, self, direction.inverted())

    def finds_observations_closest(self, direction: Direction) -> Closest:
        """
        Creates and returns a class:Closest lookup engine
        that can resolve the correct cell from this selection
        relative to any given observation.
        """

        if not self._label:
            raise MissingLabelError(
                """
                You are trying to create a lookup engine for a selection of
                cells using the .resolve_observations_closest() method but
                have not yet assigned a label to said selection of cells.

                Please use the .label_as() method to assign a label before
                attempting this.
            """
            )

        # The constructor we provide to the user advertises that the column
        # lookup engines "finds the observations" but this is purely a
        # conceptual trick to make a more user friendly api.
        # In reality that's exactly backwards, an observation actually finds
        # a column value (by being passed to the lookup engine constructed below).
        # As such we need to reverse the stated direction.
        return Closest(self.label, self, direction.inverted())

    def finds_observations_within(
        self, direction: Direction, start: Direction, end: Direction
    ) -> Within:
        """
        Creates and returns a class:Within lookup engine
        that can resolve the correct cell from this selection
        relative to any given observation.
        """

        if not self.label:
            raise MissingLabelError(
                """
                You are trying to create a lookup engine for a selection of
                cells using the .resolve_observations_within() method but have
                not yet assigned a label to said selection of cells.

                Please use the .label_as() method to assign a label before
                attempting this.
            """
            )

        return Within(
            self.label,
            self,
            direction.inverted(),
            start.inverted(),
            end.inverted(),
            table=self.name,
        )


def _explain(selectable: Selectable, comment: str):
    if selectable._explain or selectable._explain_path:
        assert len(selectable.cells) > 0, (
            f'Error: stage "EXPLAIN: {comment}" results in 0 cells selected.'
            " You cannot preview nothing."
        )
        selectable = selectable.label_as(f"EXPLAIN: {comment}")
        preview(selectable, selection_boundary=True, path=selectable._explain_path)
